import os
import re
import sys
import timeit
from pathlib import Path, PurePosixPath

import h5py
import numpy as np
import psutil
from mpi4py import MPI
from tqdm import tqdm

from .h5 import h5_print_contents
from .utils import even_print, format_time_string

# ======================================================================

def _calc_statistics_xpln(self,**kwargs):
    '''
    Calculate statistics for an unsteady volumetric measurement
        which has a small number of points in the [x] direction
    - Mean
    - Covariance
    - Triple products
    - Skewness, Kurtosis, and further standardized central moments
    - Probability distribution function (PDF)
    '''
    
    if (self.rank==0):
        verbose = True
    else:
        verbose = False
    
    if verbose: print('\n'+'rgd.calc_statistics_xpln()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    ## assert that the opened RGD has fsubtype 'unsteady' (i.e. is NOT a prime file)
    if (self.fsubtype!='unsteady'):
        raise ValueError
    
    rx = kwargs.get('rx',1)
    ry = kwargs.get('ry',1)
    rz = kwargs.get('rz',1)
    rt = kwargs.get('rt',1)
    
    # cy = kwargs.get('cy',1) ## number of subdivisions per rank [y] range
    # if not isinstance(cy,int):
    #     raise TypeError('cy should be an int')
    # if (cy<1):
    #     raise TypeError('cy should be an int')
    
    sy = kwargs.get('sy',1) ## number of [y] layers to read at a time
    if not isinstance(sy,int) or (sy<1):
        raise TypeError('sy should be a positive non-zero int')
    
    #fn_dat_stats = kwargs.get('fn_dat_stats',None) ## filename for output pickle (.dat) file
    fn_h5_out  = kwargs.get('fn_h5_out',None) ## filename for output HDF5 (.h5) file
    
    n_bins = kwargs.get('n_bins',512) ## n bins for histogram (PDF) calculation
    
    ## For now only distribute data in [y] --> allows [x,z] mean before Send/Recv
    if (rx!=1):
        raise AssertionError('rx!=1')
    if (rz!=1):
        raise AssertionError('rz!=1')
    if (rt!=1):
        raise AssertionError('rt!=1')
    
    if not isinstance(ry,int) or (ry<1):
        raise ValueError('ry should be a positive non-zero int')
    
    ## Check the choice of ranks per dimension
    if (rx*ry*rz*rt != self.n_ranks):
        raise AssertionError('rx*ry*rz*rt != self.n_ranks')
    if (rx>self.nx):
        raise AssertionError('rx>self.nx')
    if (ry>self.ny):
        raise AssertionError('ry>self.ny')
    if (rz>self.nz):
        raise AssertionError('rz>self.nz')
    if (rt>self.nt):
        raise AssertionError('rt>self.nt')
    
    if (self.ny%ry!=0):
        raise ValueError('ny not divisible by ry')
    
    ## Distribute 4D data over ranks --> here only in [y]
    ryl_ = np.array_split(np.arange(self.ny,dtype=np.int64),ry)
    ryl = [[b[0],b[-1]+1] for b in ryl_ ]
    ry1,ry2 = ryl[self.rank]
    nyr = ry2 - ry1
    
    ## Check all [y] ranges have same size
    for ryl_ in ryl:
        if not (ryl_[1]-ryl_[0]==nyr):
            raise ValueError('[y] chunks are not even in size')
    
    # ## [y] sub chunk range --> cyl = list of ranges in ry1:ry2
    # ## cy is the NUMBER of chunks for the rank sub-range
    # cyl_ = np.array_split( np.arange(ry1,ry2) , min(cy,nyr) )
    # cyl  = [[b[0],b[-1]+1] for b in cyl_ ]
    # 
    # for nyc_ in [ cyl_[1]-cyl_[0] for cyl_ in cyl ]:
    #     if (nyc_ < 1):
    #         #raise ValueError
    #         print(f'rank {self.rank:d}: sub-range is <1')
    #         self.comm.Abort(1)
    # 
    # if 1: ## assert that [y] sub-chunk ranges are correct
    #     
    #     yi = np.arange(self.ny, dtype=np.int32)
    #     
    #     local_indices = []
    #     for cyl_ in cyl:
    #         cy1, cy2 = cyl_
    #         local_indices += [ yi_ for yi_ in yi[cy1:cy2] ]
    #     
    #     G = self.comm.gather([ self.rank , local_indices ], root=0)
    #     G = self.comm.bcast(G, root=0)
    #     
    #     all_indices = []
    #     for G_ in G:
    #         all_indices += G_[1]
    #     all_indices = np.array( sorted(all_indices), dtype=np.int32 )
    #     
    #     if not np.array_equal( all_indices , yi ):
    #         raise AssertionError
    
    if (nyr%sy!=0):
        raise ValueError('nyr not divisible by sy')
    
    ## Output filename : HDF5 (.h5)
    if (fn_h5_out is None): ## automatically determine name
        fname_path = os.path.dirname(self.fname)
        fname_base = os.path.basename(self.fname)
        fname_root, fname_ext = os.path.splitext(fname_base)
        fname_root = re.findall(r'io\S+_mpi_[0-9]+', fname_root)[0]
        fname_stats_h5_base = fname_root+'_stats.h5'
        fn_h5_out = str(PurePosixPath(fname_path, fname_stats_h5_base))
    if (Path(fn_h5_out).suffix != '.h5'):
        raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' must end in .h5")
    if os.path.isfile(fn_h5_out):
        #if (os.path.getsize(fn_h5_out) > 8*1024**3):
        #    raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' exists and is >8 [GB]. exiting for your own safety.")
        if (fn_h5_out == self.fname):
            raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' cannot be same as input filename.")
    
    # ===
    
    if verbose: even_print( 'fn_h5'      , self.fname )
    if verbose: even_print( 'fn_h5_out'  , fn_h5_out  )
    if verbose: print(72*'-')
    self.comm.Barrier()
    
    ## The data dictionary to be pickled later
    data = {}
    
    ## infile
    fsize = os.path.getsize(self.fname)/1024**3
    if verbose: even_print(os.path.basename(self.fname),'%0.1f [GB]'%fsize)
    if verbose: even_print('nx',f'{self.nx:d}')
    if verbose: even_print('ny',f'{self.ny:d}')
    if verbose: even_print('nz',f'{self.nz:d}')
    if verbose: even_print('nt',f'{self.nt:d}')
    if verbose: even_print('ngp',f'{self.ngp/1e6:0.1f} [M]')
    #if verbose: even_print('cy',f'{cy:d}')
    if verbose: even_print('sy',f'{sy:d}')
    if verbose: even_print('n_ranks',f'{self.n_ranks:d}')
    #if verbose: even_print('n_threads',f'{n_threads:d}')
    if verbose: print(72*'-')
    
    ## 0D freestream scalars
    lchar   = self.lchar   ; data['lchar']   = lchar
    U_inf   = self.U_inf   ; data['U_inf']   = U_inf
    rho_inf = self.rho_inf ; data['rho_inf'] = rho_inf
    T_inf   = self.T_inf   ; data['T_inf']   = T_inf
    
    #data['M_inf'] = self.M_inf
    data['Ma'] = self.Ma
    data['Pr'] = self.Pr
    
    ## Read in 1D coordinate arrays & re-dimensionalize
    x = np.copy( self['dims/x'][()] * self.lchar )
    y = np.copy( self['dims/y'][()] * self.lchar )
    z = np.copy( self['dims/z'][()] * self.lchar )
    t = np.copy( self['dims/t'][()] * self.tchar )
    
    nx = self.nx ; data['nx'] = nx
    ny = self.ny ; data['ny'] = ny
    nz = self.nz ; data['nz'] = nz
    nt = self.nt ; data['nt'] = nt
    
    ## Assert constant Δz
    dz0 = np.diff(z)[0]
    if not np.all(np.isclose(np.diff(z), dz0, rtol=1e-6)):
        raise NotImplementedError('Δz not constant')
    dz = np.diff(z)[0]
    
    ## Dimensional [s]
    dt = self.dt * self.tchar
    np.testing.assert_allclose(dt, t[1]-t[0], rtol=1e-12, atol=1e-12)
    
    t_meas = self.duration * self.tchar
    np.testing.assert_allclose(t_meas, t.max()-t.min(), rtol=1e-12, atol=1e-12)
    
    zrange = z.max() - z.min()
    
    data['x'] = x
    data['y'] = y
    data['z'] = z
    
    data['t'] = t
    data['t_meas'] = t_meas
    data['dt'] = dt
    data['dz'] = dz
    data['zrange'] = zrange
    
    if verbose: even_print( 'Δt/tchar'       , f'{dt/self.tchar:0.8f}' )
    if verbose: even_print( 'Δt'             , f'{dt:0.3e} [s]'        )
    if verbose: even_print( 'duration/tchar' , f'{self.duration:0.1f}' )
    if verbose: even_print( 'duration'       , f'{self.duration*self.tchar:0.3e} [s]' )
    if verbose: print(72*'-')
    
    ## report
    if verbose:
        even_print('Δt'     , f'{dt    :0.5e} [s]' )
        even_print('t_meas' , f'{t_meas:0.5e} [s]' )
        even_print('Δz'     , f'{dz0   :0.5e} [m]' )
        even_print('zrange' , f'{zrange:0.5e} [m]' )
        #print(72*'-')
    
    # ===
    
    ## key    = str:scalar name
    ## value  = tuple:recipe,
    ##          bool:do_mean,
    ##          bool:do_pdf,
    ##          bool:do_skew_kurt
    ##
    ## recipe elements:
    ##    - if tuple : ( str:scalar , bool:ρ weighting ) --> always mean-removed
    ##    - if str   : str:scalar
    
    scalars_dict = {
        
        'r_uIIuIIuII' : [ ( 'rho', ('u',True ), ('u',True ), ('u',True ) ), True,True,True ], ## ρ·u″u″u″
        'uIuIuI'      : [ (        ('u',False), ('u',False), ('u',False) ), True,True,True ], ## u′u′u′
        
        'r_uIIuIIvII' : [ ( 'rho', ('u',True ), ('u',True ), ('v',True ) ), True,True,True ], ## ρ·u″u″v″
        'uIuIvI'      : [ (        ('u',False), ('u',False), ('v',False) ), True,True,True ], ## u′u′v′
        
        'r_uIIvIIvII' : [ ( 'rho', ('u',True ), ('v',True ), ('v',True ) ), True,True,True ], ## ρ·u″v″v″
        'uIvIvI'      : [ (        ('u',False), ('v',False), ('v',False) ), True,True,True ], ## u′v′v′
        
        'r_vIIvIIvII' : [ ( 'rho', ('v',True ), ('v',True ), ('v',True ) ), True,True,True ], ## ρ·v″v″v″
        'vIvIvI'      : [ (        ('v',False), ('v',False), ('v',False) ), True,True,True ], ## v′v′v′
        
        'r_vIIwIIwII' : [ ( 'rho', ('v',True ), ('w',True ), ('w',True ) ), True,True,True ], ## ρ·v″w″w″
        'vIwIwI'      : [ (        ('v',False), ('w',False), ('w',False) ), True,True,True ], ## v′w′w′
        
        'r_uIIuII' : [ ( 'rho', ('u',True ), ('u',True ) ), True,True,True ], ## ρ·u″u″
        'r_vIIvII' : [ ( 'rho', ('v',True ), ('v',True ) ), True,True,True ], ## ρ·v″v″
        'r_wIIwII' : [ ( 'rho', ('w',True ), ('w',True ) ), True,True,True ], ## ρ·w″w″
        'r_uIIvII' : [ ( 'rho', ('u',True ), ('v',True ) ), True,True,True ], ## ρ·u″v″
        'r_uIIwII' : [ ( 'rho', ('u',True ), ('w',True ) ), True,True,True ], ## ρ·u″w″
        'r_vIIwII' : [ ( 'rho', ('v',True ), ('w',True ) ), True,True,True ], ## ρ·v″w″
        
        'uIuI'     : [ (        ('u',False), ('u',False) ), True,True,True ], ## u′u′
        'vIvI'     : [ (        ('v',False), ('v',False) ), True,True,True ], ## v′v′
        'wIwI'     : [ (        ('w',False), ('w',False) ), True,True,True ], ## w′w′
        'uIvI'     : [ (        ('u',False), ('v',False) ), True,True,True ], ## u′v′
        'uIwI'     : [ (        ('u',False), ('w',False) ), True,True,True ], ## u′w′
        'vIwI'     : [ (        ('v',False), ('w',False) ), True,True,True ], ## v′w′
        
        'uITI'     : [ (        ('u',False), ('T',False) ), True,True,True ], ## u′T′
        'vITI'     : [ (        ('v',False), ('T',False) ), True,True,True ], ## v′T′
        'wITI'     : [ (        ('w',False), ('T',False) ), True,True,True ], ## w′T′
        
        'TITI'     : [ (        ('T',False), ('T',False) ), True,True,False ], ## T′T′
        'TIITII'   : [ (        ('T',True ), ('T',True ) ), True,True,False ], ## T″T″
        'r_TIITII' : [ ( 'rho', ('T',True ), ('T',True ) ), True,True,False ], ## ρ·T″T″
        
        'rho'      : [ ( 'rho',                ) ,  True, True, False, ], ## ρ
        'rhoI'     : [ (        ('rho',False), ) , False, True,  True, ], ## ρ′
        
        'T'        : [ ( 'T',                )   ,  True, True, False, ], ## T
        'TI'       : [ (        ('T',False), )   , False, True,  True, ], ## T′
        'TII'      : [ (        ('T',True ), )   , False, True,  True, ], ## T″
        'r_TII'    : [ ( 'rho', ('T',True ), )   , False, True,  True, ], ## ρ·T″
        
        'p'        : [ ( 'p',              )   ,  True, True, False, ], ## p
        'pI'       : [ (      ('p',False), )   , False, True,  True, ], ## p′
        
        'u'        : [ ( 'u',                ) ,  True, True, False, ], ## u
        'uI'       : [ (        ('u',False), ) , False, True,  True, ], ## u′
        #'uII'      : [ (        ('u',True ), ) , False, True,  True, ], ## u″
        'r_uII'    : [ ( 'rho', ('u',True ), ) , False, True,  True, ], ## ρ·u″
        
        'v'        : [ ( 'v',                ) ,  True, True, False, ], ## v
        'vI'       : [ (        ('v',False), ) , False, True,  True, ], ## v′
        'r_vII'    : [ ( 'rho', ('v',True ), ) , False, True,  True, ], ## ρ·v″
        
        'w'        : [ ( 'w',                ) ,  True, True, False, ], ## w
        'wI'       : [ (        ('w',False), ) , False, True,  True, ], ## w′
        'r_wII'    : [ ( 'rho', ('w',True ), ) , False, True,  True, ], ## ρ·w″
        
        }
    
    dtype_unsteady = np.float64
    
    scalars_avg  = []
    scalars_pdf  = []
    scalars_hos_ = []
    for s,ss in scalars_dict.items():
        recipe, do_mean, do_pdf, do_skew_kurt = ss
        if do_mean:
            scalars_avg.append(s)
        if do_pdf:
            scalars_pdf.append(s)
        if do_skew_kurt:
            scalars_hos_.append(s)
    
    scalars_hos=[]
    for s_ in ['skew','kurt','scm5','scm6']: ## 'scm' = Standardized central moment
        for ss_ in scalars_hos_:
            scalars_hos.append(f'{ss_}_{s_}')
    
    data_avg  = np.zeros(shape=(nyr,)         , dtype={'names':scalars_avg, 'formats':[ np.float64 for sss in scalars_avg ]})
    data_bins = np.zeros(shape=(nyr,n_bins+1) , dtype={'names':scalars_pdf, 'formats':[ np.float64 for sss in scalars_pdf ]})
    data_pdf  = np.zeros(shape=(nyr,n_bins)   , dtype={'names':scalars_pdf, 'formats':[ np.float64 for sss in scalars_pdf ]})
    data_hos  = np.zeros(shape=(nyr,)         , dtype={'names':scalars_hos, 'formats':[ np.float64 for sss in scalars_hos ]})
    
    # ==============================================================
    # Check memory
    # ==============================================================
    
    hostname = MPI.Get_processor_name()
    mem_free_gb = psutil.virtual_memory().free / 1024**3
    G = self.comm.gather([ self.rank , hostname , mem_free_gb ], root=0)
    G = self.comm.bcast(G, root=0)
    
    host_mem = {}
    for rank, host, mem in G:
        if host not in host_mem or mem < host_mem[host]:
            host_mem[host] = mem
    total_free = sum(host_mem.values())
    
    if verbose:
        print(72*'-')
        for key,value in host_mem.items():
            even_print(f'RAM free {key}', f'{int(np.floor(value)):d} [GB]')
        even_print('RAM free (local,min)', f'{int(np.floor(min(host_mem.values()))):d} [GB]')
        even_print('RAM free (global)', f'{int(np.floor(total_free)):d} [GB]')
    
    shape_read = (nx,sy,nz,nt) ## local
    if verbose: even_print('read shape (local)', f'[{nx:d},{sy:d},{nz:d},{nt:d}]')
    data_gb = np.dtype(np.float64).itemsize * np.prod(shape_read) / 1024**3
    if verbose: even_print('read size (global)', f'{int(np.ceil(data_gb*ry)):d} [GB]')
    
    if verbose: even_print('read size (global) ×6', f'{int(np.ceil(data_gb*ry*6)):d} [GB]')
    ram_usage_est = data_gb*ry*6/total_free
    if verbose: even_print('RAM usage estimate', f'{100*ram_usage_est:0.1f} [%]')
    
    self.comm.Barrier()
    if (ram_usage_est>0.80):
        print('RAM consumption might be too high. exiting.')
        self.comm.Abort(1)
    
    # ==============================================================
    # Main loop
    # ==============================================================
    
    self.comm.Barrier()
    if verbose:
        progress_bar = tqdm(
            #total=cy*len(scalars_dict),
            total=len(scalars_dict)*(nyr//sy),
            ncols=100,
            desc='statistics',
            leave=True,
            file=sys.stdout,
            mininterval=0.1,
            smoothing=0.,
            #bar_format="\033[B{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\033[A\n\b",
            bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
            ascii="░█",
            colour='#00ffff',
            )
    
    ## Scalar dict loop
    for s,ss in scalars_dict.items(): ## e.g. where s='r_uIIvII' & ss=[ ( 'rho', ('u',True ), ('v',True ) ), True,True,False ]
        
        if verbose: tqdm.write(72*'-')
        
        recipe, do_mean, do_pdf, do_skew_kurt = ss
        
        if verbose:
            tqdm.write(even_print('computing',s,s=True,))
        
        # if verbose: ## report per-rank read shape
        #     nyc_max = max([ cyl_[1]-cyl_[0] for cyl_ in cyl ])
        #     data_gb = 8 * nx * nyc_max * nz * nt / 1024**3
        #     tqdm.write(even_print('read shape per rank', f'[{nx:d},{nyc_max:d},{nz:d},{nt:d}] · 8 [Bytes] --> {data_gb:0.1f} [GB]', s=True))
        #     tqdm.write(even_print('mem per read', f'{self.n_ranks*data_gb:0.1f} [GB]', s=True))
        #     tqdm.write(even_print('mem per read ×6', f'{self.n_ranks*data_gb*6.:0.3f} [GB]', s=True)) ## genoa3tb64c, ?R ...
        
        ## Should ρ be read?
        read_rho = False
        for s_ in recipe:
            if isinstance(s_, str) and (s_=='rho'):
                read_rho = True
            elif isinstance(s_, str) and (s_!='rho'):
                pass
            elif isinstance(s_, tuple):
                if not isinstance(s_[1], bool):
                    raise ValueError
                if s_[1]: ## i.e. density-weighted
                    read_rho = True
            else:
                raise ValueError
        
        ## [y] loop outer (grid subchunks within rank)
        # for cyl_ in cyl:
        #     cy1, cy2 = cyl_
        #     nyc = cy2 - cy1
        
        for ci in range(nyr//sy):
            cy1 = ry1 + ci*sy
            cy2 = cy1 + sy
            nyc = cy2 - cy1
            
            ## Read ρ
            if read_rho:
                
                #rho = np.zeros(shape=(nx,nyc,nz,nt), dtype=dtype_unsteady) ## buffer... chunk range context
                
                dset = self['data/rho']
                self.comm.Barrier()
                t_start = timeit.default_timer()
                if self.usingmpi:
                    with dset.collective:
                        #rho[:,:,:,:] = dset[:,:,cy1:cy2,:].T.astype(np.float64)
                        rho = dset[:,:,cy1:cy2,:].T
                else:
                    #rho[:,:,:,:] = dset[()].T.astype(np.float64)
                    rho = dset[()].T
                self.comm.Barrier()
                t_delta = timeit.default_timer() - t_start
                data_gb = dset.dtype.itemsize * self.nx * ry * (cy2-cy1) * self.nz * self.nt / 1024**3
                if verbose:
                    tqdm.write(even_print('read: ρ', '%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
                
                ## Cast to double
                rho = rho.astype(np.float64)
                
                ## Re-dimensionalize
                rho *= self.rho_inf
                
                ## ρ mean in [t] --> leave [x,y,z,1]
                rho_avg = np.mean(rho, axis=3, dtype=np.float64, keepdims=True)
            
            else:
                rho     = None ; del rho
                rho_avg = None ; del rho_avg
            
            ## Product buffer for multiplication --> !!! notices ones() here and not zeros() !!!
            data_accum = np.ones(shape=(nx,nyc,nz,nt), dtype=dtype_unsteady) ## chunk range context
            
            ## Read unsteady scalar data, remove mean, <density weight>, multiply
            for sss in recipe:
                
                if isinstance(sss, str) and (sss=='rho'): ## ρ
                    
                    ## Multiply product accumulator, ρ was already read and is already dimensional
                    data_accum *= rho
                
                elif isinstance(sss, str) and (sss!='rho'): ## scalar which will NOT be mean-removed
                    
                    #data_X = np.zeros(shape=(nx,nyc,nz,nt), dtype=dtype_unsteady) ## buffer... chunk range context
                    dset = self[f'data/{sss}']
                    
                    ## Read
                    self.comm.Barrier()
                    t_start = timeit.default_timer()
                    if self.usingmpi:
                        with dset.collective:
                            #data_X[:,:,:,:] = dset[:,:,cy1:cy2,:].T.astype(np.float64)
                            #data_X = np.copy( dset[:,:,cy1:cy2,:].T )
                            data_X = dset[:,:,cy1:cy2,:].T
                    else:
                        #data_X[:,:,:,:] = dset[()].T.astype(np.float64)
                        #data_X = np.copy( dset[()].T )
                        data_X = dset[()].T
                    self.comm.Barrier()
                    t_delta = timeit.default_timer() - t_start
                    data_gb = dset.dtype.itemsize * self.nx * ry * (cy2-cy1) * self.nz * self.nt / 1024**3
                    if verbose:
                        tqdm.write(even_print(f'read: {sss}', '%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
                    
                    ## Cast to double
                    data_X = data_X.astype(np.float64)
                    
                    ## Re-dimensionalize
                    if sss in ['u','v','w',]:
                        data_X *= self.U_inf
                    elif sss in ['T',]:
                        data_X *= self.T_inf
                    elif sss in ['rho',]:
                        data_X *= self.rho_inf
                    elif sss in ['p',]:
                        data_X *= self.rho_inf * self.U_inf**2
                    else:
                        raise ValueError(f"condition needed for redimensionalizing '{str(sss)}'")
                    
                    ## MULTIPLY product accumulator
                    data_accum *= data_X
                
                elif isinstance(sss, tuple): ## scalar which WILL be mean-removed (with- or without ρ-weighting)
                    
                    if (len(sss)!=2):
                        raise ValueError
                    
                    sn, do_density_weighting = sss ## e.g. ('u',True)
                    
                    #data_X = np.zeros(shape=(nx,nyc,nz,nt), dtype=dtype_unsteady) ## buffer... chunk range context
                    dset = self[f'data/{sn}']
                    
                    ## Read
                    self.comm.Barrier()
                    t_start = timeit.default_timer()
                    if self.usingmpi:
                        with dset.collective:
                            #data_X[:,:,:,:] = dset[:,:,cy1:cy2,:].T.astype(np.float64)
                            #data_X = np.copy( dset[:,:,cy1:cy2,:].T )
                            data_X = dset[:,:,cy1:cy2,:].T
                    else:
                        #data_X[:,:,:,:] = dset[()].T.astype(np.float64)
                        #data_X = np.copy( dset[()].T )
                        data_X = dset[()].T
                    self.comm.Barrier()
                    t_delta = timeit.default_timer() - t_start
                    data_gb = dset.dtype.itemsize * self.nx * ry * (cy2-cy1) * self.nz * self.nt / 1024**3
                    if verbose:
                        tqdm.write(even_print(f'read: {sn}', '%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
                    
                    ## Cast to double
                    data_X = data_X.astype(np.float64)
                    
                    ## Redimensionalize
                    if sn in ['u','v','w',]:
                        data_X *= self.U_inf
                    elif sn in ['T',]:
                        data_X *= self.T_inf
                    elif sn in ['rho',]:
                        data_X *= self.rho_inf
                    elif sn in ['p',]:
                        data_X *= self.rho_inf * self.U_inf**2
                    else:
                        raise ValueError(f"condition needed for redimensionalizing '{str(sn)}'")
                    
                    ## avg(□) or avg(ρ·□)/avg(ρ)
                    if do_density_weighting:
                        data_X_mean = np.mean( rho*data_X , axis=3, dtype=np.float64, keepdims=True) ## [x,y,z,1]
                        data_X_mean /= rho_avg
                    else:
                        data_X_mean = np.mean( data_X , axis=3, dtype=np.float64, keepdims=True) ## [x,y,z,1]
                    
                    ## Reynolds prime □′ or Favre prime □″
                    data_X -= data_X_mean
                    
                    ## Assert avg(□′)==0 or avg(ρ·□″)==0
                    if do_density_weighting:
                        a_ = np.mean(rho*data_X, axis=3, dtype=np.float64, keepdims=True)
                    else:
                        a_ = np.mean(data_X, axis=3, dtype=np.float64, keepdims=True)
                    #np.testing.assert_allclose(a_, np.zeros_like(a_), atol=1e-3)
                    if not np.allclose( a_, np.zeros_like(a_), atol=1e-3 ):
                        print(f'rank {self.rank:d}: avg(□′)!=0 or avg(ρ·□″)!=0')
                        self.comm.Abort(1)
                    
                    ## MULTIPLY product accumulator by □′ | □″
                    data_accum *= data_X
                
                else:
                    raise ValueError
            
            # ===============================================================================
            # At this point you have the UNSTEADY 4D [x,y,z,t] data according to 'recipe'
            # ===============================================================================
            
            yiA = cy1 - ry1
            yiB = cy2 - ry1
            
            ## Mean in [t] --> leave [x,y,z,1]
            d_mean = np.mean( data_accum, axis=(3,), keepdims=True, dtype=np.float64)
            if ( d_mean.shape != (nx,nyc,nz,1,) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            if do_mean: ## mean in [x,z] --> leave [y]
                data_avg[s][yiA:yiB] = np.squeeze( np.mean( d_mean, axis=(0,2), dtype=np.float64) )
            
            ## [y] loop inner ([y] indices within subdivision within rank)
            for yi in range(cy1,cy2):
                
                yii  = yi - cy1 ## chunk local
                yiii = yi - ry1 ## rank local
                
                if do_pdf:
                    
                    d_ = np.copy( data_accum[:,yii,:,:].ravel() )
                    if ( d_.shape != (nx*1*nz*nt,) ):
                        print(f'rank {self.rank:d}: shape violation')
                        self.comm.Abort(1)
                    pdf_ , bin_edges_ = np.histogram( d_ , bins=n_bins , density=True )
                    data_bins[s][yiii,:] = bin_edges_
                    data_pdf[s][yiii,:]  = pdf_
                
                if do_skew_kurt: ## Calculate standardized 
                    
                    #d_ = np.copy( data_accum[:,yii,:,:] ) ## [x,1,z,t]
                    d_ = np.zeros(shape=(nx,1,nz,nt), dtype=dtype_unsteady)
                    d_[:,:,:,:] = data_accum[:,yii,:,:][:,np.newaxis,:,:] ## [x,1,z,t]
                    if ( d_.shape != (nx,1,nz,nt) ):
                        print(f'rank {self.rank:d}: shape violation')
                        self.comm.Abort(1)
                    
                    ## ## is it interpretable as a Reynolds mean-removed quantity? (□′)
                    ## interpretable_as_Re_mean = False
                    ## a_ = np.mean(d_, axis=3, dtype=np.float64, keepdims=True) ## mean in [t]
                    ## np.testing.assert_equal( a_.shape, (nx,1,nz,1,) )
                    ## if np.allclose(a_, np.zeros_like(a_), atol=1e-10):
                    ##     interpretable_as_Re_mean = True
                    ## 
                    ## ## is it interpretable as a Favre mean-removed quantity? (□″)
                    ## interpretable_as_Fv_mean = False
                    ## if read_rho and not interpretable_as_Re_mean:
                    ##     #rho_ = np.copy( rho[:,yii,:,:] )
                    ##     rho_ = np.zeros(shape=(nx,1,nz,nt), dtype=dtype_unsteady)
                    ##     rho_[:,:,:,:] = rho[:,yii,:,:][:,np.newaxis,:,:] ## [x,1,z,t]
                    ##     np.testing.assert_equal( rho_.shape, (nx,1,nz,nt) )
                    ##     
                    ##     a_ = np.mean(d_*rho_, axis=3, dtype=np.float64, keepdims=True) ## mean in [t]
                    ##     np.testing.assert_equal( a_.shape, (nx,1,nz,1,) )
                    ##     if np.allclose(a_, np.zeros_like(a_), atol=1e-10):
                    ##         interpretable_as_Fv_mean = True
                    ## 
                    ## if (not interpretable_as_Re_mean) and (not interpretable_as_Fv_mean):
                    ##     print(f'{s} has non-zero Favre and Reynolds mean, i.e. is not <□′>!=0 and <ρ□″>!=0')
                    ##     self.comm.Abort(1)
                    
                    dI = np.copy( d_.ravel() ) ## □′ or □″
                    if ( dI.shape != (nx*1*nz*nt,) ):
                        print(f'rank {self.rank:d}: shape violation')
                        self.comm.Abort(1)
                    
                    d_std = np.sqrt( np.mean( dI**2 , dtype=np.float64 ) )
                    
                    if np.isclose(d_std, 0., atol=1e-08): ## e.g. @ no-slip wall
                        d_skew = 0.
                        d_kurt = 0.
                        d_scm5 = 0.
                        d_scm6 = 0.
                    else:
                        d_skew = np.mean( dI**3 , dtype=np.float64 ) / d_std**3
                        d_kurt = np.mean( dI**4 , dtype=np.float64 ) / d_std**4
                        d_scm5 = np.mean( dI**5 , dtype=np.float64 ) / d_std**5
                        d_scm6 = np.mean( dI**6 , dtype=np.float64 ) / d_std**6
                    
                    data_hos[f'{s}_skew'][yiii] = d_skew
                    data_hos[f'{s}_kurt'][yiii] = d_kurt
                    data_hos[f'{s}_scm5'][yiii] = d_scm5
                    data_hos[f'{s}_scm6'][yiii] = d_scm6
            
            self.comm.Barrier() ## [y] loop outer (chunks within rank)
            if verbose: progress_bar.update()
            
            # break ## debug
    
    if verbose:
        progress_bar.close()
        print(72*'-')
    
    # ==============================================================
    # Write HDF5 (.h5) file
    # ==============================================================
    
    ## Open on rank 0 and write attributes, dimensions, etc.
    if (self.rank==0):
        with h5py.File(fn_h5_out, 'w') as hfw:
            
            ## Write floats,ints as top-level attributes
            for key,val in data.items():
                if isinstance(data[key], (int,np.int32,np.int64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], (float,np.float32,np.float64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], np.ndarray):
                    pass
                else:
                    print(f'key {key} is type {str(type(data[key]))}')
                    self.comm.Abort(1)
            
            ## Write numpy arrays
            hfw.create_dataset( 'dims/x' , data=x ) ## [m]
            hfw.create_dataset( 'dims/y' , data=y ) ## [m]
            hfw.create_dataset( 'dims/z' , data=z ) ## [m]
            hfw.create_dataset( 'dims/t' , data=t ) ## [s]
            
            ## Initialize datasets
            for scalar in scalars_avg:
                hfw.create_dataset( f'avg/{scalar}', shape=(ny,), dtype=np.float64, chunks=None, data=np.full((ny,),0.,np.float64) )
            for scalar in scalars_pdf:
                hfw.create_dataset( f'bins/{scalar}', shape=(ny,n_bins+1), dtype=np.float64, chunks=(1,n_bins+1), data=np.full((ny,n_bins+1),0.,np.float64) )
            for scalar in scalars_pdf:
                hfw.create_dataset( f'pdf/{scalar}', shape=(ny,n_bins), dtype=np.float64, chunks=(1,n_bins), data=np.full((ny,n_bins),0.,np.float64) )
            for scalar in scalars_hos:
                hfw.create_dataset( f'hos/{scalar}', shape=(ny,), dtype=np.float64, chunks=None, data=np.full((ny,),0.,np.float64) )
    
    self.comm.Barrier()
    
    with h5py.File(fn_h5_out, 'a', driver='mpio', comm=self.comm) as hfw:
        
        ## Collectively write avg,bins,pdf,hos
        for scalar in scalars_avg:
            dset = hfw[f'avg/{scalar}']
            with dset.collective:
                dset[ry1:ry2] = data_avg[scalar][:]
        for scalar in scalars_pdf:
            dset = hfw[f'bins/{scalar}']
            with dset.collective:
                dset[ry1:ry2,:] = data_bins[scalar][:,:]
        for scalar in scalars_pdf:
            dset = hfw[f'pdf/{scalar}']
            with dset.collective:
                dset[ry1:ry2,:] = data_pdf[scalar][:,:]
        for scalar in scalars_hos:
            dset = hfw[f'hos/{scalar}']
            with dset.collective:
                dset[ry1:ry2] = data_hos[scalar][:]
    
    ## Report file contents
    self.comm.Barrier()
    if (self.rank==0):
        even_print( os.path.basename(fn_h5_out) , f'{(os.path.getsize(fn_h5_out)/1024**2):0.1f} [MB]' )
        print(72*'-')
        with h5py.File(fn_h5_out,'r') as hfr:
            h5_print_contents(hfr)
    self.comm.Barrier()
    
    if verbose: print(72*'-')
    if verbose: print('total time : rgd.calc_statistics_xpln() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return



import math
import os
import sys
import timeit

import numpy as np
import psutil
from tqdm import tqdm

from .gradient import gradient
from .h5 import h5_chunk_sizer, h5_ds_force_allocate_chunks
from .utils import even_print, format_time_string

# ======================================================================

def _calc_lambda2(self, **kwargs):
    '''
    Calculate λ₂ & Q
    Jeong & Hussain (1996) : https://doi.org/10.1017/S0022112095000462
    '''
    
    if (self.rank==0):
        verbose = True
    else:
        verbose = False
    
    rx = kwargs.get('rx',1)
    ry = kwargs.get('ry',1)
    rz = kwargs.get('rz',1)
    
    save_Q       = kwargs.get('save_Q',True)
    save_lambda2 = kwargs.get('save_lambda2',True)
    
    acc          = kwargs.get('acc',4)
    edge_stencil = kwargs.get('edge_stencil','half')
    
    chunk_kb         = kwargs.get('chunk_kb',4*1024) ## h5 chunk size: default 4 [MB]
    chunk_constraint = kwargs.get('chunk_constraint',(1,None,None,None)) ## the 'constraint' parameter for sizing h5 chunks
    chunk_base       = kwargs.get('chunk_base',2)
    
    d = 1 ## derivative order
    stencil_npts = 2*math.floor((d+1)/2) - 1 + acc
    
    if ((stencil_npts-1)%2 != 0):
        raise AssertionError
    
    stencil_npts_one_side = int( (stencil_npts-1)/2 )
    
    # ===
    
    if verbose: print('\n'+'turbx.rgd.calc_lambda2()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    ## checks
    if all([(save_Q is False),(save_lambda2 is False)]):
        raise AssertionError('neither λ-2 nor Q set to be solved')
    if not (self.open_mode=='a') or (self.open_mode=='w') or (self.open_mode=='r+'):
        raise ValueError('not able to write to hdf5 file')
    if 'data/u' not in self:
        raise ValueError('data/u not in hdf5')
    if 'data/v' not in self:
        raise ValueError('data/v not in hdf5')
    if 'data/w' not in self:
        raise ValueError('data/w not in hdf5')
    
    if (rx*ry*rz != self.n_ranks):
        raise AssertionError('rx*ry*rz != self.n_ranks')
    if (rx>self.nx):
        raise AssertionError('rx>self.nx')
    if (ry>self.ny):
        raise AssertionError('ry>self.ny')
    if (rz>self.nz):
        raise AssertionError('rz>self.nz')
    
    if verbose: even_print('save_Q','%s'%save_Q)
    if verbose: even_print('save_lambda2','%s'%save_lambda2)
    if verbose: even_print('rx','%i'%rx)
    if verbose: even_print('ry','%i'%ry)
    if verbose: even_print('rz','%i'%rz)
    if verbose: print(72*'-')
    
    #t_read = 0.
    #t_write = 0.
    #data_gb_read = 0.
    #data_gb_write = 0.
    #t_q_crit = 0.
    #t_l2     = 0.
    
    ## get size of infile
    fsize = os.path.getsize(self.fname)/1024**3
    if verbose: even_print(os.path.basename(self.fname),'%0.1f [GB]'%fsize)
    if verbose: even_print('nx','%i'%self.nx)
    if verbose: even_print('ny','%i'%self.ny)
    if verbose: even_print('nz','%i'%self.nz)
    if verbose: even_print('ngp','%0.1f [M]'%(self.ngp/1e6,))
    if verbose: print(72*'-')
    
    ## report memory
    mem_total_gb = psutil.virtual_memory().total/1024**3
    mem_avail_gb = psutil.virtual_memory().available/1024**3
    mem_free_gb  = psutil.virtual_memory().free/1024**3
    if verbose: even_print('mem total', '%0.1f [GB]'%mem_total_gb)
    if verbose: even_print('mem available', '%0.1f [GB]'%mem_avail_gb)
    if verbose: even_print('mem free', '%0.1f [GB]'%mem_free_gb)
    if verbose: print(72*'-')
    
    # === The 'standard' non-abutting / non-overlapping index split
    
    if self.usingmpi:
        
        comm4d = self.comm.Create_cart(dims=[rx,ry,rz], periods=[False,False,False], reorder=False)
        t4d = comm4d.Get_coords(self.rank)
        
        rxl_ = np.array_split(np.arange(self.nx,dtype=np.int64),min(rx,self.nx))
        ryl_ = np.array_split(np.arange(self.ny,dtype=np.int64),min(ry,self.ny))
        rzl_ = np.array_split(np.arange(self.nz,dtype=np.int64),min(rz,self.nz))
        #rtl_ = np.array_split(np.arange(self.nt,dtype=np.int64),min(rt,self.nt))
        
        rxl = [[b[0],b[-1]+1] for b in rxl_ ]
        ryl = [[b[0],b[-1]+1] for b in ryl_ ]
        rzl = [[b[0],b[-1]+1] for b in rzl_ ]
        #rtl = [[b[0],b[-1]+1] for b in rtl_ ]
        
        rx1, rx2 = rxl[t4d[0]]; nxr = rx2 - rx1
        ry1, ry2 = ryl[t4d[1]]; nyr = ry2 - ry1
        rz1, rz2 = rzl[t4d[2]]; nzr = rz2 - rz1
        #rt1, rt2 = rtl[t4d[3]]; ntr = rt2 - rt1
    
    else:
        
        nxr = self.nx
        nyr = self.ny
        nzr = self.nz
        #ntr = self.nt
    
    ## Check rank / grid distribution
    if self.usingmpi and False:
        for ri in range(self.n_ranks):
            self.comm.Barrier()
            if (self.rank == ri):
                print('rank %04d : rx1=%i rx2=%i ry1=%i ry2=%i rz1=%i rz2=%i'%(self.rank, rx1,rx2, ry1,ry2, rz1,rz2))
                sys.stdout.flush()
    
    if self.usingmpi: self.comm.Barrier()
    
    # === Extend the rank ranges (spatial range overlap)
    
    if self.usingmpi:
        
        n_overlap = stencil_npts_one_side + 3
        
        xA = 0
        xB = nxr
        yA = 0
        yB = nyr
        zA = 0
        zB = nzr
        
        ## backup non-overlapped bounds
        rx1_orig, rx2_orig = rx1, rx2
        ry1_orig, ry2_orig = ry1, ry2
        rz1_orig, rz2_orig = rz1, rz2
        
        ## overlap in [x]
        if (t4d[0]!=0):
            rx1, rx2 = rx1-n_overlap, rx2
            xA += n_overlap
            xB += n_overlap
        if (t4d[0]!=rx-1):
            rx1, rx2 = rx1, rx2+n_overlap
        
        ## overlap in [y]
        if (t4d[1]!=0):
            ry1, ry2 = ry1-n_overlap, ry2
            yA += n_overlap
            yB += n_overlap
        if (t4d[1]!=ry-1):
            ry1, ry2 = ry1, ry2+n_overlap
        
        ## overlap in [z]
        if (t4d[2]!=0):
            rz1, rz2 = rz1-n_overlap, rz2
            zA += n_overlap
            zB += n_overlap
        if (t4d[2]!=rz-1):
            rz1, rz2 = rz1, rz2+n_overlap
        
        ## update (rank local) nx,ny,nz
        nxr = rx2 - rx1
        nyr = ry2 - ry1
        nzr = rz2 - rz1
    
    ## Check rank / grid distribution
    if self.usingmpi and False:
        for ri in range(self.n_ranks):
            self.comm.Barrier()
            if (self.rank == ri):
                print('rank %04d : rx1=%i rx2=%i ry1=%i ry2=%i rz1=%i rz2=%i'%(self.rank, rx1,rx2, ry1,ry2, rz1,rz2))
                sys.stdout.flush()
    
    if self.usingmpi: self.comm.Barrier()
    
    # ===
    
    dtype = self['data/u'].dtype
    #itemsize = dtype.itemsize
    float_bytes = dtype.itemsize
    
    data_gb = float_bytes*self.nx*self.ny*self.nz*self.nt / 1024**3
    shape  = (self.nt,self.nz,self.ny,self.nx)
    chunks = h5_chunk_sizer(nxi=shape, constraint=chunk_constraint, size_kb=chunk_kb, base=chunk_base, itemsize=float_bytes)
    
    # === Initialize 4D arrays in HDF5
    
    if save_lambda2:
        
        self.scalars_dtypes_dict['lambda2'] = dtype
        
        if verbose:
            even_print('initializing data/lambda2','%0.2f [GB]'%(data_gb,))
        if ('data/lambda2' in self):
            del self['data/lambda2']
        dset = self.create_dataset(
                            'data/lambda2', 
                            shape=shape, 
                            dtype=dtype,
                            chunks=chunks,
                            )
        
        if self.usingmpi: self.comm.Barrier()
        
        ## write dummy data
        if self.rank==0:
            h5_ds_force_allocate_chunks(dset,verbose=verbose)
        
        if self.usingmpi: self.comm.Barrier()
        
        chunk_kb_ = np.prod(dset.chunks)*float_bytes / 1024. ## actual
        if verbose:
            even_print('chunk shape (t,z,y,x)','%s'%str(dset.chunks))
            even_print('chunk size','%i [KB]'%int(round(chunk_kb_)))
    
    if save_Q:
        
        self.scalars_dtypes_dict['Q'] = dtype
        
        if verbose:
            even_print('initializing data/Q','%0.2f [GB]'%(data_gb,))
        if ('data/Q' in self):
            del self['data/Q']
        dset = self.create_dataset(
                            'data/Q', 
                            shape=shape, 
                            dtype=dtype,
                            chunks=chunks,
                            )
        
        if self.usingmpi: self.comm.Barrier()
        
        ## write dummy data
        if self.rank==0:
            h5_ds_force_allocate_chunks(dset,verbose=verbose)
        
        if self.usingmpi: self.comm.Barrier()
        
        chunk_kb_ = np.prod(dset.chunks)*float_bytes / 1024. ## actual
        if verbose:
            even_print('chunk shape (t,z,y,x)', str(dset.chunks))
            even_print('chunk size', f'{int(round(chunk_kb_)):d} [KB]')
    
    if verbose: print(72*'-')
    
    # ===
    
    if verbose:
        progress_bar = tqdm(total=self.nt, ncols=100, desc='calc λ2', leave=False, file=sys.stdout)
    
    for ti in self.ti:
        
        # === Read u,v,w
        
        dset = self['data/u']
        dtype = dset.dtype
        float_bytes = dtype.itemsize
        if self.usingmpi: self.comm.Barrier()
        t_start = timeit.default_timer()
        if self.usingmpi:
            with dset.collective:
                u_ = dset[ti,rz1:rz2,ry1:ry2,rx1:rx2].T
        else:
            u_ = dset[ti,:,:,:].T
        if self.usingmpi: self.comm.Barrier()
        t_delta = timeit.default_timer() - t_start
        data_gb = float_bytes * self.nx * self.ny * self.nz * 1 / 1024**3
        if verbose:
            tqdm.write( even_print('read u', '%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True ) )
        
        dset = self['data/v']
        dtype = dset.dtype
        float_bytes = dtype.itemsize
        if self.usingmpi: self.comm.Barrier()
        t_start = timeit.default_timer()
        if self.usingmpi:
            with dset.collective:
                v_ = dset[ti,rz1:rz2,ry1:ry2,rx1:rx2].T
        else:
            v_ = dset[ti,:,:,:].T
        if self.usingmpi: self.comm.Barrier()
        t_delta = timeit.default_timer() - t_start
        data_gb = float_bytes * self.nx * self.ny * self.nz * 1 / 1024**3
        if verbose:
            tqdm.write( even_print('read v', '%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True ) )
        
        dset = self['data/w']
        dtype = dset.dtype
        float_bytes = dtype.itemsize
        if self.usingmpi: self.comm.Barrier()
        t_start = timeit.default_timer()
        if self.usingmpi:
            with dset.collective:
                w_ = dset[ti,rz1:rz2,ry1:ry2,rx1:rx2].T
        else:
            w_ = dset[ti,:,:,:].T
        if self.usingmpi: self.comm.Barrier()
        t_delta = timeit.default_timer() - t_start
        data_gb = float_bytes * self.nx * self.ny * self.nz * 1 / 1024**3
        if verbose:
            tqdm.write( even_print('read w', '%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True ) )
        
        # ===
        
        if self.usingmpi:
            x_ = np.copy(self.x[rx1:rx2])
            y_ = np.copy(self.y[ry1:ry2])
            z_ = np.copy(self.z[rz1:rz2])
        else:
            x_ = np.copy(self.x)
            y_ = np.copy(self.y)
            z_ = np.copy(self.z)
        
        # === ∂(u)/∂(x,y,z)
        
        ddx_u = gradient(u_, x_, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
        ddy_u = gradient(u_, y_, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        ddz_u = gradient(u_, z_, axis=2, acc=acc, edge_stencil=edge_stencil, d=1)
        
        # === ∂(v)/∂(x,y,z)
        
        ddx_v = gradient(v_, x_, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
        ddy_v = gradient(v_, y_, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        ddz_v = gradient(v_, z_, axis=2, acc=acc, edge_stencil=edge_stencil, d=1)
        
        # === ∂(w)/∂(x,y,z)
        
        ddx_w = gradient(w_, x_, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
        ddy_w = gradient(w_, y_, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
        ddz_w = gradient(w_, z_, axis=2, acc=acc, edge_stencil=edge_stencil, d=1)
        
        ## Free memory
        u_ = None; del u_
        v_ = None; del v_
        w_ = None; del w_
        
        strain = np.copy( np.stack((np.stack((ddx_u, ddy_u, ddz_u), axis=3),
                                    np.stack((ddx_v, ddy_v, ddz_v), axis=3),
                                    np.stack((ddx_w, ddy_w, ddz_w), axis=3)), axis=4) )
        
        t_delta = timeit.default_timer() - t_start
        if verbose: tqdm.write( even_print('get strain ∂(u,v,v)/∂(x,y,z)' , '%0.3f [s]'%(t_delta,), s=True) )
        
        ## Free memory
        ddx_u = None; del ddx_u
        ddy_u = None; del ddy_u
        ddz_u = None; del ddz_u
        ddx_v = None; del ddx_v
        ddy_v = None; del ddy_v
        ddz_v = None; del ddz_v
        ddx_w = None; del ddx_w
        ddy_w = None; del ddy_w
        ddz_w = None; del ddz_w
        
        # === Get the rate-of-strain & vorticity tensors
        
        S = np.copy( 0.5*(strain + np.transpose(strain, axes=(0,1,2,4,3))) ) ## strain rate tensor (symmetric)
        Ω = np.copy( 0.5*(strain - np.transpose(strain, axes=(0,1,2,4,3))) ) ## rotation rate tensor (anti-symmetric)
        # np.testing.assert_allclose(S+Ω, strain, atol=1.e-6)
        
        ## Free memory
        strain = None; del strain
        
        # === Q : second invariant of characteristics equation: λ³ + Pλ² + Qλ + R = 0
        
        if save_Q:
            
            t_start = timeit.default_timer()
            
            O_norm  = np.linalg.norm(Ω, ord='fro', axis=(3,4)) ## Frobenius norm
            S_norm  = np.linalg.norm(S, ord='fro', axis=(3,4))
            Q       = 0.5*(O_norm**2 - S_norm**2)
            
            t_delta = timeit.default_timer() - t_start
            if verbose: tqdm.write(even_print('calc Q','%s'%format_time_string(t_delta), s=True))
            
            dset = self['data/Q']
            dtype = dset.dtype
            float_bytes = dtype.itemsize
            if self.usingmpi: self.comm.Barrier()
            t_start = timeit.default_timer()
            if self.usingmpi:
                with dset.collective:
                    dset[ti,rz1_orig:rz2_orig,ry1_orig:ry2_orig,rx1_orig:rx2_orig] = Q[xA:xB,yA:yB,zA:zB].T
            else:
                dset[ti,:,:,:] = Q.T
            if self.usingmpi: self.comm.Barrier()
            t_delta = timeit.default_timer() - t_start
            data_gb = float_bytes * self.nx * self.ny * self.nz / 1024**3
            
            if verbose: tqdm.write(even_print('write Q','%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
            
            # # === second invariant : Q --> an equivalent formulation using eigenvalues (but much slower)
            # if False:
            #     Q_bak = np.copy(Q)
            #     t_start = timeit.default_timer()
            #     eigvals = np.linalg.eigvals(strain)
            #     P       = -1*np.sum(eigvals, axis=-1) ## first invariant : P
            #     SijSji  = np.einsum('xyzij,xyzji->xyz', S, S)
            #     OijOji  = np.einsum('xyzij,xyzji->xyz', Ω, Ω)
            #     Q       = 0.5*(P**2 - SijSji - OijOji)
            #     t_delta = timeit.default_timer() - t_start
            #     if verbose: tqdm.write(even_print('calc Q','%s'%format_time_string(t_delta), s=True))
            #     np.testing.assert_allclose(Q.imag, np.zeros_like(Q.imag, dtype=np.float32), atol=1e-6)
            #     Q = np.copy(Q.real)
            #     np.testing.assert_allclose(Q, Q_bak, rtol=1e-2, atol=1e-5)
            
            ## Free memory
            O_norm = None; del O_norm
            S_norm = None; del S_norm
            Q = None; del Q
        
        # === λ₂
        
        if save_lambda2:
            
            t_start = timeit.default_timer()
            
            # === S² and Ω²
            SikSkj = np.einsum('xyzik,xyzkj->xyzij', S, S)
            OikOkj = np.einsum('xyzik,xyzkj->xyzij', Ω, Ω)
            #np.testing.assert_allclose(np.matmul(S,S), SikSkj, atol=1e-6)
            #np.testing.assert_allclose(np.matmul(Ω,Ω), OikOkj, atol=1e-6)
            
            ## Free memory
            S = None; del S
            Ω = None; del Ω
            
            # === Eigenvalues of (S²+Ω²) --> a real symmetric (Hermitian) matrix
            eigvals            = np.linalg.eigvalsh(SikSkj+OikOkj, UPLO='L')
            #eigvals_sort_order = np.argsort(np.abs(eigvals), axis=3) ## sort order of λ --> magnitude (wrong)
            eigvals_sort_order = np.argsort(eigvals, axis=3) ## sort order of λ
            eigvals_sorted     = np.take_along_axis(eigvals, eigvals_sort_order, axis=3) ## do λ sort
            lambda2            = np.squeeze(eigvals_sorted[:,:,:,1]) ## λ2 is the second eigenvalue (index=1)
            t_delta            = timeit.default_timer() - t_start
            
            if verbose: tqdm.write(even_print('calc λ2','%s'%format_time_string(t_delta), s=True))
            
            dset = self['data/lambda2']
            dtype = dset.dtype
            float_bytes = dtype.itemsize
            if self.usingmpi: self.comm.Barrier()
            t_start = timeit.default_timer()
            if self.usingmpi:
                with dset.collective:
                    dset[ti,rz1_orig:rz2_orig,ry1_orig:ry2_orig,rx1_orig:rx2_orig] = lambda2[xA:xB,yA:yB,zA:zB].T
            else:
                dset[ti,:,:,:] = lambda2.T
            if self.usingmpi: self.comm.Barrier()
            t_delta = timeit.default_timer() - t_start
            data_gb = float_bytes * self.nx * self.ny * self.nz / 1024**3
            
            if verbose: tqdm.write(even_print('write λ2','%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
            
            ## Free memory
            lambda2 = None; del lambda2
            eigvals = None; del eigvals
            eigvals_sort_order = None; del eigvals_sort_order
            eigvals_sorted = None; del eigvals_sorted
        
        if verbose: progress_bar.update()
        if verbose and (ti<self.nt-1): tqdm.write( '---' )
    if verbose: progress_bar.close()
    if verbose: print(72*'-')
    
    # ===
    
    self.get_header(verbose=False)
    if verbose: even_print(self.fname, '%0.2f [GB]'%(os.path.getsize(self.fname)/1024**3))
    
    if verbose: print(72*'-')
    if verbose: print('total time : turbx.rgd.calc_lambda2() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt
import numpy as np
from scipy.integrate import cumulative_trapezoid, trapezoid
from scipy.optimize import least_squares

from turbx import (
    Blasius_solution,
    calc_d1,
    calc_d2,
    calc_d3,
    calc_d99_1d,
    calc_profile_edge_1d,
    gradient,
)

'''
========================================================================
TODO: Make this into a test!
========================================================================
'''

# main()
# ======================================================================

if __name__ == '__main__':
    
    # === Get η99 of Blasius solution
    
    def __f_opt_eta99(eta99):
        _, fp, _ = Blasius_solution(eta99)
        root = fp - 0.99
        return root
    
    sol = least_squares(
        __f_opt_eta99,
        x0=[5.0],
        bounds=(0.,20.),
        xtol=1e-11,
        gtol=1e-11,
        method='trf',
        )
    eta99 = float(sol.x[0])
    print(f'η99 = {eta99:0.14f}') ## 4.90998951330668
    
    # === Make Blasius data
    
    n        = 200
    eta      = np.linspace(0,20,n) ## η
    #eta      = np.linspace(0,3,n) ## η
    y_ov_d99 = np.copy( eta / eta99 )
    
    # ===
    
    f, fp, fpp = Blasius_solution(eta)
    
    ## Plot
    plt.close('all')
    plt.rcParams.update({'font.size': 9})
    fig1 = plt.figure(figsize=(5.31/2,4), dpi=200)
    ax1 = plt.gca()
    ax1.plot(fp,y_ov_d99,c='red',lw=0.6)
    #ax1.set_ylim(0,1.5)
    ax1.set_ylim(bottom=0)
    ax1.set_ylabel(r'$y/\delta_{99}$')
    ax1.set_xlim(0,1.05)
    ax1.set_xlabel(r'$u/U_\infty$')
    fig1.tight_layout(pad=0.25)
    #plt.show()
    
    # === Dimensionalize data & get BL parameters
    
    U_inf = 5.     ## m/s
    nu    = 1.5e-5 ## m^2/s
    x     = 200.   ## m
    Re_x  = U_inf * x / nu
    
    y = np.copy( eta * np.sqrt(nu * x / U_inf) )
    u = U_inf * fp
    
    c0     = 0.3320573362149271
    u_tau  = U_inf * np.sqrt(c0) * Re_x**(-1/4)
    d99    = eta99 * np.sqrt(nu * x / U_inf)
    dnu    = nu / u_tau
    Re_tau = u_tau * d99 / nu
    print(f'Reτ = {Re_tau:0.14f}')
    np.testing.assert_allclose(Re_tau, eta99*np.sqrt(c0)*Re_x**(1/4), rtol=1e-11)
    np.testing.assert_allclose(Re_tau, d99/dnu, rtol=1e-11)
    
    uplus = np.copy( u / u_tau )
    yplus = np.copy( y / dnu   )
    
    uplus_inf = U_inf / u_tau
    cf        = 2/uplus_inf**2
    print(f'1000*cf = {1000*cf:0.14f}')
    
    # ===
    
    # def kernel_d1(eta):
    #     _,fp,_ = Blasius_solution(eta)
    #     return 1. - fp
    # 
    # def kernel_d2(eta):
    #     _,fp,_ = Blasius_solution(eta)
    #     return fp * (1. - fp)
    # 
    # def kernel_d3(eta):
    #     _,fp,_ = Blasius_solution(eta)
    #     return fp * (1. - fp*fp)
    
    # ## δ1 = c1 * sqrt(nu*x/U_inf)
    # ## δ2 = c2 * sqrt(nu*x/U_inf)
    # ## δ3 = c3 * sqrt(nu*x/U_inf)
    # from scipy.integrate import quad
    # c1, err = quad(kernel_d1, 0., 50., epsabs=1e-9, epsrel=1e-9, limit=int(1e3))
    # c2, err = quad(kernel_d2, 0., 50., epsabs=1e-9, epsrel=1e-9, limit=int(1e3))
    # c3, err = quad(kernel_d3, 0., 50., epsabs=1e-9, epsrel=1e-9, limit=int(1e3))
    # print(f'c1={c1:0.12f}')
    # print(f'c2={c2:0.12f}')
    # print(f'c3={c3:0.12f}')
    
    c1 = 1.720787657522
    c2 = 0.664114672431
    c3 = 1.044375475238
    
    ## Analytical δn
    d1 = c1 * np.sqrt(nu*x/U_inf)
    d2 = c2 * np.sqrt(nu*x/U_inf)
    d3 = c3 * np.sqrt(nu*x/U_inf)
    
    H12 = d1/d2
    H32 = d3/d2
    
    ## Assert: analytical δn vs numerical δn
    ## - these are grid-dependent
    np.testing.assert_allclose(d1, calc_d1(y=y,u=u), rtol=(1/n))
    np.testing.assert_allclose(d2, calc_d2(y=y,u=u), rtol=(1/n))
    np.testing.assert_allclose(d3, calc_d3(y=y,u=u), rtol=(1/n))
    
    ## Assert: calc_d1(), numerical only, using u_edge=u[-1]
    ## - should be identical to machine tolerance
    np.testing.assert_allclose(
        trapezoid(y=1.-(u/u[-1]), x=y),
        calc_d1(y=y,u=u),
        rtol=1e-11,
        )
    
    ## Assert: calc_d1(), numerical only, using u_edge=U_inf
    ## - should be identical to machine tolerance
    np.testing.assert_allclose(
        trapezoid(y=1.-(u/U_inf), x=y),
        calc_d1(y=y,u=u,u_edge=U_inf),
        rtol=1e-11,
        )
    
    # ==================================================================
    # Determine an edge (don't just use last index)
    # ==================================================================
    
    ddy_u = gradient(u=u,x=y,acc=6,edge_stencil='full')
    
    y_edge,j_edge = calc_profile_edge_1d(y=y, ddy_u=ddy_u, epsilon=1e-6, check_eps=False)
    
    ## Assert: calc_d1(), numerical only, using u_edge=U_inf
    ## - should be identical to machine tolerance
    np.testing.assert_allclose(
        cumulative_trapezoid(y=1.-(u/U_inf), x=y, initial=0.)[j_edge],
        calc_d1(y=y,u=u,u_edge=U_inf,j_edge=j_edge),
        rtol=1e-11,
        )
    
    ## Assert: calc_d2(), numerical only, using u_edge=U_inf
    ## - should be identical to machine tolerance
    np.testing.assert_allclose(
        cumulative_trapezoid(y=(u/U_inf)*(1.-(u/U_inf)), x=y, initial=0.)[j_edge],
        calc_d2(y=y,u=u,u_edge=U_inf,j_edge=j_edge),
        rtol=1e-11,
        )
    
    ## Assert: calc_d3(), numerical only, using u_edge=U_inf
    ## - should be identical to machine tolerance
    np.testing.assert_allclose(
        cumulative_trapezoid(y=(u/U_inf)*(1.-(u/U_inf)**2), x=y, initial=0.)[j_edge],
        calc_d3(y=y,u=u,u_edge=U_inf,j_edge=j_edge),
        rtol=1e-11,
        )
    
    # ==================================================================
    # δ99
    # ==================================================================
    
    ## Assert: analytical δ99 vs numerical δ99
    np.testing.assert_allclose(
        d99,
        calc_d99_1d(y=y, u=u, u_edge=U_inf),
        rtol=1e-5,
        )
    
    ## Assert: analytical δ99 vs numerical δ99
    np.testing.assert_allclose(
        d99,
        calc_d99_1d(y=y, u=u, u_edge=U_inf, j_edge=j_edge),
        rtol=1e-5,
        )



import os

import h5py
import numpy as np
from mpi4py import MPI

from .utils import even_print

# ======================================================================

class eas4(h5py.File):
    '''
    Interface class for EAS4 files
    ------------------------------
    - super()'ed h5py.File class
    - EAS4 is the HDF5-based output format from the flow solver NS3D
    - 3D dataset storage ([x,y,z] per [t])
    '''
    
    def __init__(self, *args, **kwargs):
        
        ## if grid MUST be read as 3D, i.e. the GMODE=(5,5,5), only read if this is TRUE
        ## this can lead to huge RAM usage in MPI mode, so OFF by default
        ##
        ## if a single dimension has GMODE=5 but not ALL, i.e. (5,5,2), this allows for
        ## grid to be read as some combination of 2D&1D
        self.read_3d_grid = kwargs.pop('read_3d_grid', False)
        
        self.fname, openMode = args
        
        self.fname_path = os.path.dirname(self.fname)
        self.fname_base = os.path.basename(self.fname)
        self.fname_root, self.fname_ext = os.path.splitext(self.fname_base)
        
        ## default to libver='latest' if none provided
        if ('libver' not in kwargs):
            kwargs['libver'] = 'latest'
        
        if (openMode!='r'):
            raise ValueError('turbx.eas4(): opening EAS4 in anything but read mode \'r\' is not allowed!')
        
        ## catch possible user error --> user tries to open non-EAS4 with turbx.eas4()
        if (self.fname_ext!='.eas'):
            raise ValueError('turbx.eas4() should not be used to open non-EAS4 files')
        
        ## determine if using mpi
        if ('driver' in kwargs) and (kwargs['driver']=='mpio'):
            if ('comm' not in kwargs):
                raise ValueError("if driver='mpio', then comm should be provided")
            self.usingmpi = True
        else:
            self.usingmpi = False
        
        ## determine communicator & rank info
        if self.usingmpi:
            self.comm    = kwargs['comm']
            self.n_ranks = self.comm.Get_size()
            self.rank    = self.comm.Get_rank()
        else:
            self.comm    = None
            self.n_ranks = 1
            self.rank    = 0
            if ('comm' in kwargs):
                del kwargs['comm']
        
        ## set library version to latest (if not otherwise set)
        if ('libver' not in kwargs):
            kwargs['libver']='latest'
        
        ## unique kwargs (not h5py.File kwargs) --> pop() rather than get()
        no_indep_rw = kwargs.pop('no_indep_rw', False)
        if not isinstance(no_indep_rw, bool):
            raise ValueError('no_indep_rw must be type bool')
        
        ## set MPI hints, passed through 'mpi_info' dict
        if self.usingmpi:
            if ('info' in kwargs):
                self.mpi_info = kwargs['info']
            else:
                mpi_info = MPI.Info.Create()
                ##
                mpi_info.Set('romio_ds_write' , 'disable'   )
                mpi_info.Set('romio_ds_read'  , 'disable'   )
                #mpi_info.Set('romio_cb_read'  , 'automatic' )
                #mpi_info.Set('romio_cb_write' , 'automatic' )
                mpi_info.Set('romio_cb_read'  , 'enable' )
                mpi_info.Set('romio_cb_write' , 'enable' )
                
                ## ROMIO -- collective buffer size
                mpi_info.Set('cb_buffer_size' , str(int(round(1024**3))) ) ## 1 [GB]
                
                ## ROMIO -- force collective I/O
                if no_indep_rw:
                    mpi_info.Set('romio_no_indep_rw' , 'true' )
                
                ## ROMIO -- N Aggregators
                #mpi_info.Set('cb_nodes' , str(min(16,self.n_ranks//2)) )
                mpi_info.Set('cb_nodes' , str(min(16,self.n_ranks)) )
                
                kwargs['info'] = mpi_info
                self.mpi_info = mpi_info
        
        # === HDF5 tuning factors (independent of MPI I/O driver)
        
        ## rdcc_w0 : preemption policy (weight) for HDF5's raw data chunk cache
        ## - influences how HDF5 evicts chunks from the per-process chunk cache
        ## - 1.0 favors retaining fully-read chunks (good for read-heavy access)
        ## - 0.0 favors recently-used chunks (better for partial writes)
        if ('rdcc_w0' not in kwargs):
            kwargs['rdcc_w0'] = 0.75
        
        ## rdcc_nbytes : maximum total size of the HDF5 raw chunk cache per dataset per process
        if ('rdcc_nbytes' not in kwargs):
            kwargs['rdcc_nbytes'] = int(1*1024**3) ## 1 [GB]
        
        ## rdcc_nslots : number of hash table slots in the raw data chunk cache
        ## - should be ~= ( rdcc_nbytes / chunk size )
        if ('rdcc_nslots' not in kwargs):
            #kwargs['rdcc_nslots'] = 16381 ## prime
            kwargs['rdcc_nslots'] = kwargs['rdcc_nbytes'] // (2*1024**2) ## assume 2 [MB] chunks
            #kwargs['rdcc_nslots'] = kwargs['rdcc_nbytes'] // (128*1024**2) ## assume 128 [MB] chunks
        
        self.domainName = 'DOMAIN_000000' ## turbx only handles one domain for now
        
        ## eas4() unique kwargs (not h5py.File kwargs) --> pop() rather than get()
        verbose = kwargs.pop('verbose',False)
        self.verbose = verbose
        ## force = kwargs.pop('force',False) ## --> dont need, always read-only!
        
        ## call actual h5py.File.__init__()
        super(eas4, self).__init__(*args, **kwargs)
        self.get_header()
    
    def get_header(self):
        
        EAS4_NO_G   = 1
        EAS4_X0DX_G = 2
        EAS4_ALL_G  = 4
        #EAS4_FULL_G = 5
        gmode_dict  = {1:'EAS4_NO_G', 2:'EAS4_X0DX_G', 3:'EAS4_UDEF_G', 4:'EAS4_ALL_G', 5:'EAS4_FULL_G'}
        
        self.bform = self['Kennsatz'].attrs['bform'] ## binary format : 1=single  , 2=double
        self.dform = self['Kennsatz'].attrs['dform'] ## data format   : 1=C-order , 2=Fortran-order
        
        # === Characteristic values
        
        if self.verbose: print(72*'-')
        Ma    = self['Kennsatz']['FLOWFIELD_PROPERTIES'].attrs['Ma'][0]
        Re    = self['Kennsatz']['FLOWFIELD_PROPERTIES'].attrs['Re'][0]
        Pr    = self['Kennsatz']['FLOWFIELD_PROPERTIES'].attrs['Pr'][0]
        kappa = self['Kennsatz']['FLOWFIELD_PROPERTIES'].attrs['kappa'][0]
        R     = self['Kennsatz']['FLOWFIELD_PROPERTIES'].attrs['R'][0]
        p_inf = self['Kennsatz']['FLOWFIELD_PROPERTIES'].attrs['p_inf'][0]
        T_inf = self['Kennsatz']['FLOWFIELD_PROPERTIES'].attrs['T_inf'][0]
        
        ## !!! import what is called 'C_Suth' in NS3D as 'S_Suth' !!!
        S_Suth = self['Kennsatz']['VISCOUS_PROPERTIES'].attrs['C_Suth'][0]
        
        mu_Suth_ref = self['Kennsatz']['VISCOUS_PROPERTIES'].attrs['mu_Suth_ref'][0]
        T_Suth_ref  = self['Kennsatz']['VISCOUS_PROPERTIES'].attrs['T_Suth_ref'][0]
        
        C_Suth = mu_Suth_ref/(T_Suth_ref**(3/2))*(T_Suth_ref + S_Suth) ## [kg/(m·s·√K)]
        
        if self.verbose: even_print('Ma'          , '%0.2f [-]'           % Ma          )
        if self.verbose: even_print('Re'          , '%0.1f [-]'           % Re          )
        if self.verbose: even_print('Pr'          , '%0.3f [-]'           % Pr          )
        if self.verbose: even_print('T_inf'       , '%0.3f [K]'           % T_inf       )
        if self.verbose: even_print('p_inf'       , '%0.1f [Pa]'          % p_inf       )
        if self.verbose: even_print('kappa'       , '%0.3f [-]'           % kappa       )
        if self.verbose: even_print('R'           , '%0.3f [J/(kg·K)]'    % R           )
        if self.verbose: even_print('mu_Suth_ref' , '%0.6E [kg/(m·s)]'    % mu_Suth_ref )
        if self.verbose: even_print('T_Suth_ref'  , '%0.2f [K]'           % T_Suth_ref  )
        if self.verbose: even_print('S_Suth'      , '%0.2f [K]'           % S_Suth      )
        if self.verbose: even_print('C_Suth'      , '%0.5e [kg/(m·s·√K)]' % C_Suth      ) ## actually derived
        
        # === Characteristic values : derived
        
        # mu_inf_1 = 14.58e-7*T_inf**1.5/(T_inf+110.4)
        # mu_inf_2 = mu_Suth_ref*(T_inf/T_Suth_ref)**(3/2) * ((T_Suth_ref+S_Suth)/(T_inf+S_Suth))
        # mu_inf_3 = C_Suth*T_inf**(3/2)/(T_inf+S_Suth)
        # if not np.isclose(mu_inf_1, mu_inf_2, rtol=1e-14):
        #     raise AssertionError('inconsistency in Sutherland calc --> check')
        # if not np.isclose(mu_inf_2, mu_inf_3, rtol=1e-14):
        #     raise AssertionError('inconsistency in Sutherland calc --> check')
        # mu_inf    = mu_inf_3
        
        if self.verbose: print(72*'-')
        mu_inf      = mu_Suth_ref*(T_inf/T_Suth_ref)**(3/2) * ((T_Suth_ref+S_Suth)/(T_inf+S_Suth))
        rho_inf     = p_inf / ( R * T_inf )
        nu_inf      = mu_inf/rho_inf
        a_inf       = np.sqrt(kappa*R*T_inf)
        U_inf       = Ma*a_inf
        cp          = R*kappa/(kappa-1.)
        cv          = cp/kappa
        recov_fac   = Pr**(1/3) ## for turbulent boundary layer
        Taw         = T_inf + recov_fac*U_inf**2/(2*cp)
        lchar       = Re * nu_inf / U_inf
        tchar       = lchar / U_inf
        uchar       = U_inf ## alias
        M_inf       = Ma ## alias
        p_tot_inf   = p_inf   * (1 + (kappa-1)/2 * M_inf**2)**(kappa/(kappa-1))
        T_tot_inf   = T_inf   * (1 + (kappa-1)/2 * M_inf**2)
        rho_tot_inf = rho_inf * (1 + (kappa-1)/2 * M_inf**2)**(1/(kappa-1))
        
        if self.verbose: even_print('rho_inf'         , '%0.3f [kg/m³]'    % rho_inf   )
        if self.verbose: even_print('mu_inf'          , '%0.6E [kg/(m·s)]' % mu_inf    )
        if self.verbose: even_print('nu_inf'          , '%0.6E [m²/s]'     % nu_inf    )
        if self.verbose: even_print('a_inf'           , '%0.6f [m/s]'      % a_inf     )
        if self.verbose: even_print('U_inf'           , '%0.6f [m/s]'      % U_inf     )
        if self.verbose: even_print('cp'              , '%0.3f [J/(kg·K)]' % cp        )
        if self.verbose: even_print('cv'              , '%0.3f [J/(kg·K)]' % cv        )
        #if self.verbose: even_print('recovery factor' , '%0.6f [-]'        % recov_fac )
        if self.verbose: even_print('Taw'             , '%0.3f [K]'        % Taw       )
        if self.verbose: even_print('lchar'           , '%0.6E [m]'        % lchar     )
        if self.verbose: even_print('tchar'           , '%0.6E [s]'        % tchar     )
        if self.verbose: print(72*'-'+'\n')
        
        # ===
        
        self.Ma           = Ma
        self.Re           = Re
        self.Pr           = Pr
        self.kappa        = kappa
        self.R            = R
        self.p_inf        = p_inf
        self.T_inf        = T_inf
        self.C_Suth       = C_Suth
        self.S_Suth       = S_Suth
        self.mu_Suth_ref  = mu_Suth_ref
        self.T_Suth_ref   = T_Suth_ref
        
        self.rho_inf     = rho_inf
        self.mu_inf      = mu_inf
        self.nu_inf      = nu_inf
        self.a_inf       = a_inf
        self.U_inf       = U_inf
        self.cp          = cp
        self.cv          = cv
        self.recov_fac   = recov_fac
        self.Taw         = Taw
        self.lchar       = lchar
        
        self.tchar       = tchar
        self.uchar       = uchar
        self.M_inf       = M_inf
        self.p_tot_inf   = p_tot_inf
        self.T_tot_inf   = T_tot_inf
        self.rho_tot_inf = rho_tot_inf
        
        # === Check if this a 2D average file like 'mean_flow_mpi.eas'
        
        if self.verbose: print(72*'-')
        if ('/Kennsatz/AUXILIARY/AVERAGING' in self):
            self.total_avg_time       = self['/Kennsatz/AUXILIARY/AVERAGING'].attrs['total_avg_time'][0]
            self.total_avg_iter_count = self['/Kennsatz/AUXILIARY/AVERAGING'].attrs['total_avg_iter_count'][0]
            if self.verbose: even_print('total_avg_time', '%0.2f'%self.total_avg_time)
            if self.verbose: even_print('total_avg_iter_count', '%i'%self.total_avg_iter_count)
            self.measType = 'mean'
        else:
            self.measType = 'unsteady'
        if self.verbose: even_print('meas type', '\'%s\''%self.measType)
        if self.verbose: print(72*'-'+'\n')
        
        # === Grid info
        
        ndim1 = int( self['Kennsatz/GEOMETRY/%s'%self.domainName].attrs['DOMAIN_SIZE'][0] )
        ndim2 = int( self['Kennsatz/GEOMETRY/%s'%self.domainName].attrs['DOMAIN_SIZE'][1] )
        ndim3 = int( self['Kennsatz/GEOMETRY/%s'%self.domainName].attrs['DOMAIN_SIZE'][2] )
        
        nx = self.nx = int(ndim1)
        ny = self.ny = int(ndim2)
        nz = self.nz = int(ndim3)
        
        if (self.measType=='mean'):
            nz = self.nz = 1
        
        ngp = self.ngp = nx*ny*nz
        
        if self.verbose: print('grid info\n'+72*'-')
        if self.verbose: even_print('nx',  '%i'%nx  )
        if self.verbose: even_print('ny',  '%i'%ny  )
        if self.verbose: even_print('nz',  '%i'%nz  )
        if self.verbose: even_print('ngp', '%i'%ngp )
        
        gmode_dim1 = self.gmode_dim1 = self['Kennsatz/GEOMETRY/%s'%self.domainName].attrs['DOMAIN_GMODE'][0]
        gmode_dim2 = self.gmode_dim2 = self['Kennsatz/GEOMETRY/%s'%self.domainName].attrs['DOMAIN_GMODE'][1]
        gmode_dim3 = self.gmode_dim3 = self['Kennsatz/GEOMETRY/%s'%self.domainName].attrs['DOMAIN_GMODE'][2]
        
        if self.verbose: even_print( 'gmode dim1' , '%i / %s'%(gmode_dim1,gmode_dict[gmode_dim1]) )
        if self.verbose: even_print( 'gmode dim2' , '%i / %s'%(gmode_dim2,gmode_dict[gmode_dim2]) )
        if self.verbose: even_print( 'gmode dim3' , '%i / %s'%(gmode_dim3,gmode_dict[gmode_dim3]) )
        if self.verbose: print(72*'-')
        
        # === Read grid
        
        ## dset object handles, no data is read yet
        dset1 = self[f'Kennsatz/GEOMETRY/{self.domainName}/dim01']
        dset2 = self[f'Kennsatz/GEOMETRY/{self.domainName}/dim02']
        dset3 = self[f'Kennsatz/GEOMETRY/{self.domainName}/dim03']
        
        if any([(gmode_dim1==5), (gmode_dim2==5), (gmode_dim3==5)]):
            self.is_curvilinear = True
            self.is_rectilinear = False
        else:
            self.is_curvilinear = False
            self.is_rectilinear = True
        
        dim1_data = None
        dim2_data = None
        dim3_data = None
        
        ## Read grid data if ND<3 or if 'read_3d_grid' opt is set
        if dset1.ndim<3 or self.read_3d_grid:
            dim1_data = np.copy(dset1[()])
        if dset2.ndim<3 or self.read_3d_grid:
            dim2_data = np.copy(dset2[()])
        if dset3.ndim<3 or self.read_3d_grid:
            dim3_data = np.copy(dset3[()])
        
        # === Expand gmodes 1,2 --> 4 (always ok, even for MPI mode)
        
        ## Convert EAS4_NO_G to EAS4_ALL_G (1 --> 4) --> always do this
        ## dim_n_data are already numpy arrays of shape (1,) --> no conversion necessary, just update 'gmode_dimn' attr
        if (gmode_dim1 == EAS4_NO_G):
            gmode_dim1 = self.gmode_dim1 = EAS4_ALL_G
        if (gmode_dim2 == EAS4_NO_G):
            gmode_dim2 = self.gmode_dim2 = EAS4_ALL_G
        if (gmode_dim3 == EAS4_NO_G):
            gmode_dim3 = self.gmode_dim3 = EAS4_ALL_G
        
        ## Convert EAS4_X0DX_G to EAS4_ALL_G (2 --> 4) --> always do this
        if (gmode_dim1 == EAS4_X0DX_G):
            dim1_data  = np.linspace(dim1_data[0],dim1_data[0]+dim1_data[1]*(ndim1-1), ndim1, dtype=np.float64)
            gmode_dim1 = self.gmode_dim1 = EAS4_ALL_G
        if (gmode_dim2 == EAS4_X0DX_G):
            dim2_data  = np.linspace(dim2_data[0],dim2_data[0]+dim2_data[1]*(ndim2-1), ndim2, dtype=np.float64)
            gmode_dim2 = self.gmode_dim2 = EAS4_ALL_G
        if (gmode_dim3 == EAS4_X0DX_G):
            dim3_data  = np.linspace(dim3_data[0],dim3_data[0]+dim3_data[1]*(ndim3-1), ndim3, dtype=np.float64)
            gmode_dim3 = self.gmode_dim3 = EAS4_ALL_G
        
        # ===
        
        x = self.x = np.copy(dim1_data)
        y = self.y = np.copy(dim2_data)
        z = self.z = np.copy(dim3_data)
        
        if self.verbose: even_print('x_min', '%0.2f'%x.min())
        if self.verbose: even_print('x_max', '%0.2f'%x.max())
        if self.is_rectilinear:
            if (self.nx>2):
                if self.verbose: even_print('dx begin : end', '%0.3E : %0.3E'%( (x[1]-x[0]), (x[-1]-x[-2]) ))
        if self.verbose: even_print('y_min', '%0.2f'%y.min())
        if self.verbose: even_print('y_max', '%0.2f'%y.max())
        if self.is_rectilinear:
            if (self.ny>2):
                if self.verbose: even_print('dy begin : end', '%0.3E : %0.3E'%( (y[1]-y[0]), (y[-1]-y[-2]) ))
        if self.verbose: even_print('z_min', '%0.2f'%z.min())
        if self.verbose: even_print('z_max', '%0.2f'%z.max())
        if self.is_rectilinear:
            if (self.nz>2):
                if self.verbose: even_print('dz begin : end', '%0.3E : %0.3E'%( (z[1]-z[0]), (z[-1]-z[-2]) ))
        if self.verbose: print(72*'-'+'\n')
        
        # === Time & scalar info
        
        if self.verbose: print('time & scalar info\n'+72*'-')
        
        n_scalars = self['Kennsatz/PARAMETER'].attrs['PARAMETER_SIZE'][0]
        
        if ('Kennsatz/PARAMETER/PARAMETERS_ATTRS' in self):
            scalars =  [ s.decode('utf-8').strip() for s in self['Kennsatz/PARAMETER/PARAMETERS_ATTRS'][()] ]
        else:
            ## old EAS4 structure
            scalars = [ self['Kennsatz/PARAMETER'].attrs['PARAMETERS_ATTR_%06d'%i][0].decode('utf-8').strip() for i in range(n_scalars) ]
        
        scalar_n_map = dict(zip(scalars, range(n_scalars)))
        
        self.scalars_dtypes = []
        for scalar in scalars:
            dset_path = 'Data/%s/ts_%06d/par_%06d'%(self.domainName,0,scalar_n_map[scalar])
            if (dset_path in self):
                self.scalars_dtypes.append(self[dset_path].dtype)
            else:
                #self.scalars_dtypes.append(np.float32)
                raise AssertionError('dset not found: %s'%dset_path)
        
        self.scalars_dtypes_dict = dict(zip(scalars, self.scalars_dtypes)) ## dict {<<scalar>>: <<dtype>>}
        
        nt         = self['Kennsatz/TIMESTEP'].attrs['TIMESTEP_SIZE'][0] 
        gmode_time = self['Kennsatz/TIMESTEP'].attrs['TIMESTEP_MODE'][0]
        
        ## Baseflow will not have a TIMEGRID
        if ('Kennsatz/TIMESTEP/TIMEGRID' in self):
            t = self['Kennsatz/TIMESTEP/TIMEGRID'][:]
        else:
            t = np.array( [0.] , dtype=np.float64 )
        
        if (gmode_time==EAS4_X0DX_G): ## =2 --> i.e. more than one timestep
            t = np.linspace(t[0],t[0]+t[1]*(nt - 1), nt  )
            gmode_time = EAS4_ALL_G
        else:
            #print('gmode_time : '+str(gmode_time))
            pass
        
        if t.size>1:
            dt       = float( t[1] - t[0] )
            duration = float( t[-1] - t[0] )
        else:
            dt = 0.
            duration = 0.
        
        if self.verbose: even_print('nt', '%i'%nt )
        if self.verbose: even_print('dt', '%0.6f'%dt)
        if self.verbose: even_print('duration', '%0.2f'%duration )
        
        # === Attach to instance
        
        self.n_scalars    = n_scalars
        self.scalars      = scalars
        self.scalar_n_map = scalar_n_map
        self.t            = t
        self.dt           = dt
        self.nt           = nt
        self.duration     = duration
        
        self.ti = np.arange(self.nt, dtype=np.float64)
        
        if self.verbose: print(72*'-'+'\n')
        
        ## Pack & attach a udef dict as attr for convenience
        self.udef = {
            'Ma':self.Ma,
            'Re':self.Re,
            'Pr':self.Pr,
            'kappa':self.kappa,
            'R':self.R,
            'p_inf':self.p_inf,
            'T_inf':self.T_inf,
            'S_Suth':self.S_Suth,
            'mu_Suth_ref':self.mu_Suth_ref,
            'T_Suth_ref':self.T_Suth_ref,
            
            'C_Suth':self.C_Suth,
            'mu_inf':self.mu_inf,
            'rho_inf':self.rho_inf,
            'nu_inf':self.nu_inf,
            'a_inf':self.a_inf,
            'U_inf':self.U_inf,
            'cp':self.cp,
            'cv':self.cv,
            'recov_fac':self.recov_fac,
            'Taw':self.Taw,
            'lchar':self.lchar,
            'tchar':self.tchar,
            
            'uchar':self.uchar,
            'M_inf':self.M_inf,
            
            'p_tot_inf':self.p_tot_inf,
            'T_tot_inf':self.T_tot_inf,
            'rho_tot_inf':self.rho_tot_inf,
            }
        
        return
    
    def get_mean(self, **kwargs):
        '''
        Return spanwise mean of 2D EAS4 file
        - usually each [z] represents [z]-averaged rank data
        '''
        axis = kwargs.get('axis',2) ## Default: avg over [z]
        
        if (self.measType!='mean'):
            raise NotImplementedError('get_mean() not yet valid for measType=\'%s\''%self.measType)
        
        ## Numpy structured array
        data_mean = np.zeros(shape=(self.nx,self.ny), dtype={'names':self.scalars, 'formats':self.scalars_dtypes})
        
        for si, scalar in enumerate(self.scalars):
            scalar_dtype = self.scalars_dtypes[si]
            dset_path = 'Data/%s/ts_%06d/par_%06d'%(self.domainName,0,self.scalar_n_map[scalar])
            
            if self.dform==1: ## EAS4 has C-order
                data = np.copy(self[dset_path][()])
            elif self.dform==2: ## EAS4 has Fortran-order; perform transpose
                data = np.copy(self[dset_path][()].T)
            else:
                raise ValueError(f".dform={self.dform:d} not supported.")
            
            ## Perform np.mean() with float64 accumulator!
            scalar_mean = np.mean(data, axis=axis, dtype=np.float64).astype(scalar_dtype)
            data_mean[scalar] = scalar_mean
        
        return data_mean
    
    def make_xdmf(self):
        print('make_xdmf() not yet implemented for turbx class EAS4')
        raise NotImplementedError
        return



import os
import sys
import timeit

import h5py
import numpy as np
from tqdm import tqdm

from .eas4 import eas4
from .gradient import gradient
from .h5 import h5_ds_force_allocate_chunks, h5_print_contents
from .utils import even_print, format_time_string

# ======================================================================

def _init_from_eas4_wall(self, fn_eas4, **kwargs):
    '''
    Initialize 'wall' SPD from an EAS4
    '''
    verbose = kwargs.get('verbose',True)
    if (self.rank!=0):
        verbose=False
    
    if verbose: print('\n'+'spd.init_from_eas4_wall()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    #if (self.rank==0):
    with eas4(fn_eas4,'r',comm=self.comm,driver=self.driver) as hfr:
        
        if (hfr.x.ndim!=1):
            raise RuntimeError('hfr.x.ndim!=1')
        if (hfr.z.ndim!=1):
            raise RuntimeError('hfr.z.ndim!=1')
        
        ## Primary attributes
        ats = ['Ma','Re','Pr','kappa','R','p_inf','T_inf','S_Suth','mu_Suth_ref','T_Suth_ref',]
        for at in ats:
            self.attrs[at] = hfr.udef[at]
        
        ## Derived attributes
        ats = ['C_Suth','mu_inf','rho_inf','nu_inf','a_inf','U_inf','cp','cv','lchar','tchar','uchar','M_inf',]
        for at in ats:
            if at in hfr.attrs:
                self.attrs[at] = hfr.attrs[at]
        
        ## 3D polydata grid coordinates : shape (nx,nz,3)
        xyz = np.zeros((hfr.nx,hfr.nz,3), dtype=np.float64)
        xyz[:,:,0] = hfr.x[:,np.newaxis]
        xyz[:,:,1] = 0.
        xyz[:,:,2] = hfr.z[np.newaxis,:]
        
        self.create_dataset(
            name='dims/xyz',
            chunks=(1,hfr.nz,1), ## (x,z,3)
            #shape=(hfr.nx, hfr.nz, 3),
            #dtype=np.float64,
            data=xyz,
            )
        
        self.flush()
        xyz = None; del xyz
        
        ## 1D [x],[y],[z] vector
        self.create_dataset('dims/x', data=hfr.x, chunks=None)
        self.create_dataset('dims/z', data=hfr.z, chunks=None)
        self.create_dataset('dims/y', data=np.array([0.,],dtype=np.float64), chunks=None)
        
        ## Add attributes
        self.attrs['ni'] = hfr.nx
        self.attrs['nj'] = hfr.nz
        self.attrs['n_quads'] = ( hfr.nx - 1 ) * ( hfr.nz - 1 )
        self.attrs['n_pts'] = hfr.nx * hfr.nz
        #self.attrs['nt'] = hfr.nt
        self.attrs['nt'] = 0 ## overwritten later
    
    if self.usingmpi:
        self.comm.Barrier()
    
    self.get_header(verbose=True)
    if verbose: print(72*'-')
    
    if (self.rank==0):
        with h5py.File(self.fname,'r') as hf:
            h5_print_contents(hf)
    if self.usingmpi:
        self.comm.Barrier()
    
    ## Report
    if verbose:
        print(72*'-')
        even_print( os.path.basename(self.fname), f'{(os.path.getsize(self.fname)/1024**3):0.1f} [GB]')
    if verbose: print(72*'-')
    if verbose: print('total time : spd.init_from_eas4_wall() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return

def _import_eas4_wall(self, fn_eas4_list, **kwargs):
    '''
    Directly import to 'wall' SPD file from EAS4s
    - only parallelize in [x]
    '''
    
    if (self.rank!=0):
        verbose=False
    else:
        verbose=True
    
    if verbose: print('\n'+'spd.import_eas4_wall()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    ## dont actually copy over data, just initialize datasets with 0's
    init_dsets_only = kwargs.get('init_dsets_only',False)
    
    ## import from EAS4 as 3D unchunked datasets
    ## - have to convert to 4D later
    ## - necessary workaround for disk space with 'huge' datasets
    threeD   = kwargs.get('threeD',False)
    fn_h5_3D = kwargs.get('fn_h5_3D',None)
    fi_min   = kwargs.get('fi_min',0) ## minimum 'file' ID (for restarting MOVE)
    
    if fn_h5_3D is not None:
        if not os.path.isfile(fn_h5_3D):
            raise FileNotFoundError(fn_h5_3D)
    
    # if not self.usingmpi:
    #     if verbose: print('this function has not been implemented in non-MPI mode')
    #     sys.exit(1)
    
    # if not h5py.h5.get_config().mpi:
    #     if verbose: print('h5py was not compiled for parallel usage! exiting.')
    #     sys.exit(1)
    
    acc          = kwargs.get('acc',4)
    edge_stencil = kwargs.get('edge_stencil','full')
    chunks       = kwargs.get('chunks',None)
    
    if chunks is None:
        #chunks = (4, self.nj, 50) ## (x,z,t)
        chunks = (1,self.nj,1) ## (x,z,t) --> probably a bad default!
        print("WARNING! Not providing 'chunk' could result in very bad performance")
    if not isinstance(chunks,tuple):
        raise ValueError("'chunks' must be tuple")
    if len(chunks)!=3:
        raise ValueError("len(chunks)!=3")
    
    ## delete EAS4s after import --> DANGER!
    ## ... only actually deletes files if 'do_delete.txt' is present
    delete_after_import = kwargs.get('delete_after_import',False)
    
    ## check that the passed list of EAS4 files is OK
    if not isinstance(fn_eas4_list, list):
        raise ValueError("'fn_eas4_list' must be list")
    #if not hasattr(fn_eas4_list, '__iter__'):
    #    raise ValueError("'fn_eas4_list' must be iterable")
    for fn_eas4 in fn_eas4_list:
        if not os.path.isfile(fn_eas4):
            raise FileNotFoundError(fn_eas4)
    
    ## n ranks per direction
    rx = kwargs.get('rx',1)
    if (rx != self.n_ranks):
        raise AssertionError('rx != self.n_ranks')
    
    ## nx = ni
    if not hasattr(self,'ni'):
        raise ValueError('attribute ni not found.')
    else:
        self.nx = self.ni
    
    ## nz = nj
    if not hasattr(self,'nj'):
        raise ValueError('attribute nj not found.')
    else:
        self.nz = self.nj
    
    ## distribute in [x]
    if self.usingmpi:
        rxl_ = np.array_split(np.arange(self.nx,dtype=np.int64),min(rx,self.nx))
        rxl  = [[b[0],b[-1]+1] for b in rxl_ ]
        rx1, rx2 = rxl[self.rank]
        nxr = rx2 - rx1
    else:
        rx1,rx2 = 0,self.nx
        nxr = self.nx
    
    ## Get all time info & check
    # ==============================================================
    
    if len(fn_eas4_list) == 0:
        if not hasattr(self,'t') or not hasattr(self,'nt'):
            raise RuntimeError("if fn_eas4_list=[], then t should exist")
        nt = self.nt
        t  = np.copy(self.t)
    
    else:
        if (self.rank==0):
            t = np.array([], dtype=np.float64)
            for fn_eas4 in fn_eas4_list:
                with eas4(fn_eas4, 'r', verbose=False) as hf_eas4:
                    t_ = np.copy(hf_eas4.t)
                t = np.concatenate((t,t_))
        else:
            t = np.array([], dtype=np.float64) ## 't' must exist on all ranks before bcast
        
        if self.usingmpi:
            self.comm.Barrier()
            t = self.comm.bcast(t, root=0)
        nt = t.shape[0]
        self.nt = nt
        self.t = t
        assert self.t.dtype == np.float64, f'Array dtype is {str(t.dtype)}, expected np.float64'
    
    if len(fn_eas4_list) > 0:
        if verbose: even_print('n EAS4 files','%i'%len(fn_eas4_list))
        if verbose: even_print('nt all files','%i'%nt)
        if verbose: even_print('delete after import',str(delete_after_import))
    else:
        if verbose: even_print('nt',f'{nt:d}')
    if verbose: print(72*'-')
    
    # ==============================================================
    
    ## check [t] & Δt
    if (nt>1):
        
        ## check no zero distance elements
        if (np.diff(t).size - np.count_nonzero(np.diff(t))) != 0.:
            raise AssertionError('t arr has zero-distance elements')
        else:
            if verbose: even_print('check: Δt!=0','passed')
        
        ## check monotonically increasing
        if not np.all(np.diff(t) > 0.):
            raise AssertionError('t arr not monotonically increasing')
        else:
            if verbose: even_print('check: t mono increasing','passed')
        
        ## check constant Δt
        dt0 = np.diff(t)[0]
        if not np.all(np.isclose(np.diff(t), dt0, rtol=1e-3)):
            if (self.rank==0):
                print(np.diff(t))
            #raise AssertionError('t arr not uniformly spaced')
            if verbose:
                even_print('check: constant Δt','failed')
                print('WARNING: [t] not uniformly spaced!!')
        else:
            if verbose: even_print('check: constant Δt','passed')
        
        ## (over)write [t]
        if len(fn_eas4_list) > 0:
            if self.usingmpi: self.comm.Barrier()
            if 'dims/t' in self:
                del self['dims/t']
            self.create_dataset('dims/t', data=t, chunks=None, dtype=np.float64)
            self.attrs['nt'] = nt
            self.attrs['dt'] = dt0
            self.attrs['duration'] = dt0 * (nt - 1)
            self.flush()
            if self.usingmpi: self.comm.Barrier()
            if verbose: print(72*'-')
    
    else:
        return ## nothing to do
    
    # ==============================================================
    # Initialize datasets (harmless if they exist)
    # ==============================================================
    
    if not threeD: ## ...because '3D' mode doesnt require pre-initialization
        
        scalars = [
            'tau_uy','tau_vy','tau_wy',
            'u_tau','v_tau','w_tau',
            'T','mu','nu','rho','p',
            ]
        
        dtype = np.dtype(np.float32)
        shape = (self.nx, self.nz, self.nt)
        data_gb = np.prod(shape) * dtype.itemsize / 1024**3
        
        if verbose:
            progress_bar = tqdm(
                total=len(scalars),
                ncols=100,
                desc='initialize dsets',
                leave=True,
                file=sys.stdout,
                mininterval=0.1,
                smoothing=0.,
                #bar_format="\033[B{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\033[A\n\b",
                bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
                ascii="░█",
                colour='#FF6600',
                )
        
        ## Initialize output datasets
        for scalar in scalars:
            
            dsn = f'data/{scalar}'
            if dsn not in self:
                if self.usingmpi: self.comm.Barrier()
                t_start = timeit.default_timer()
                
                if verbose:
                    tqdm.write( even_print(f'initializing data/{scalar}', f'{str(shape)} / {data_gb:0.1f} [GB]', s=True) )
                
                dset = self.create_dataset(
                                        dsn,
                                        shape=shape,
                                        dtype=dtype,
                                        chunks=chunks,
                                        )
                
                ## write dummy data to dataset to ensure that it is truly initialized
                if not self.usingmpi:
                    h5_ds_force_allocate_chunks(dset,verbose=verbose) #,quick=True)
                
                if self.usingmpi: self.comm.Barrier()
                t_delta = timeit.default_timer() - t_start
                if verbose:
                    tqdm.write( even_print(f'initialize data/{scalar}', f'{data_gb:0.2f} [GB]  {t_delta:0.2f} [s]  {(data_gb/t_delta):0.3f} [GB/s]', s=True) )
                
                chunk_kb_ = np.prod(dset.chunks) * dset.dtype.itemsize / 1024. ## actual
                if verbose:
                    tqdm.write( even_print('chunk shape (x,z,t)', str(dset.chunks), s=True) )
                    tqdm.write( even_print('chunk size', f'{int(round(chunk_kb_)):d} [KB]', s=True) )
            
            if verbose: progress_bar.update()
        
        if self.usingmpi:
            self.comm.Barrier()
        if verbose:
            progress_bar.close()
            print(72*'-')
    
    # ==============================================================
    # Read,process,write (from EAS4)
    # ==============================================================
    
    if not init_dsets_only and fn_h5_3D is None:
        
        if verbose:
            progress_bar = tqdm(
                total=len(fn_eas4_list),
                ncols=100,
                desc='import',
                leave=True,
                file=sys.stdout,
                mininterval=0.1,
                smoothing=0.,
                #bar_format="\033[B{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\033[A\n\b",
                bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
                ascii="░█",
                colour='#FF6600',
                )
        
        tii = 0 ## timestep index counter full series
        
        for i_eas4,fn_eas4 in enumerate(fn_eas4_list):
            
            with eas4(fn_eas4, 'r', verbose=False, driver=self.driver, comm=self.comm) as hf_eas4:
                
                ## dimensional coordinates
                #x = np.copy( hf_eas4.x * hf_eas4.lchar )
                y = np.copy( hf_eas4.y * hf_eas4.lchar )
                #z = np.copy( hf_eas4.z * hf_eas4.lchar )
                
                nx = hf_eas4.nx
                ny = hf_eas4.ny
                nz = hf_eas4.nz
                nt = hf_eas4.nt ## OVERWRITING ABOVE!!!
                
                if verbose:
                    tqdm.write(even_print(os.path.basename(fn_eas4), f'{(os.path.getsize(fn_eas4)/1024**3):0.1f} [GB]', s=True))
                
                ## rank-local data read buffer for this EAS4
                scalars = ['rho','u','v','w','T','p']
                #formats = [ np.dtype(np.float32) for s in scalars ]
                formats = [ np.dtype(np.float64) for s in scalars ] ## do arithmetic in double
                data    = np.zeros(shape=(nxr,ny,nz,nt), dtype={'names':scalars, 'formats':formats})
                
                ## read (FULL file)
                self.comm.Barrier()
                t_start = timeit.default_timer()
                for ti in range(hf_eas4.nt): ## timesteps in EAS4
                    for scalar in scalars:
                        dset_path = f'Data/DOMAIN_000000/ts_{ti:06d}/par_{hf_eas4.scalar_n_map[scalar]:06d}'
                        dset = hf_eas4[dset_path]
                        with dset.collective:
                            if hf_eas4.dform==1:
                                data[scalar][:,:,:,ti] = dset[rx1:rx2,:,:]
                            elif hf_eas4.dform==2:
                                data[scalar][:,:,:,ti] = dset[:,:,rx1:rx2].T
                            else:
                                raise RuntimeError
                
                self.comm.Barrier()
                t_delta = timeit.default_timer() - t_start
                
                data_gb = os.path.getsize(fn_eas4)/1024**3
                if verbose:
                    #msg = even_print('read {os.path.basename(fn_eas4)}', f'{data_gb:0.2f} [GB]  {t_delta:0.2f} [s]  {(data_gb/t_delta):0.3f} [GB/s]', s=True)
                    msg = even_print('read', f'{data_gb:0.2f} [GB]  {t_delta:0.2f} [s]  {(data_gb/t_delta):0.3f} [GB/s]', s=True)
                    tqdm.write(msg)
                
                ## re-dimensionalize EAS4 data
                data['rho'][:,:,:,:] *= hf_eas4.rho_inf
                data['u'][:,:,:,:]   *= hf_eas4.U_inf
                data['v'][:,:,:,:]   *= hf_eas4.U_inf
                data['w'][:,:,:,:]   *= hf_eas4.U_inf
                data['T'][:,:,:,:]   *= hf_eas4.T_inf
                data['p'][:,:,:,:]   *= hf_eas4.rho_inf * hf_eas4.U_inf**2
                
                ## calculate μ and ν (dimensional)
                mu = np.zeros(shape=(nxr,ny,nz,nt), dtype=np.float64)
                nu = np.zeros(shape=(nxr,ny,nz,nt), dtype=np.float64)
                mu[:,:,:,:] = hf_eas4.mu_Suth_ref * ( data['T'][:,:,:,:] / hf_eas4.T_Suth_ref )**(3/2) * ( ( hf_eas4.T_Suth_ref + hf_eas4.S_Suth ) / ( data['T'][:,:,:,:] + hf_eas4.S_Suth ) )
                nu[:,:,:,:] = mu / data['rho'][:,:,:,:]
            
            # ==========================================================
            
            ## Dimensional wall strains
            ddy_u      = gradient( data['u'][:,:,:,:], y, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_v      = gradient( data['v'][:,:,:,:], y, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_w      = gradient( data['w'][:,:,:,:], y, axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
            ddy_u_wall = np.copy( ddy_u[:,0,:,:] )
            ddy_v_wall = np.copy( ddy_v[:,0,:,:] )
            ddy_w_wall = np.copy( ddy_w[:,0,:,:] )
            
            if (ddy_u_wall.ndim!=3): ## (x,z,t)
                raise ValueError
            if (ddy_u_wall.shape!=(nxr,nz,nt)): ## (x,z,t)
                raise ValueError
            
            ## Dimensional wall quantities
            rho_wall = np.copy( data['rho'][:,0,:,:] )
            mu_wall  = np.copy( mu[:,0,:,:]          )
            nu_wall  = np.copy( nu[:,0,:,:]          )
            T_wall   = np.copy( data['T'][:,0,:,:]   )
            p_wall   = np.copy( data['p'][:,0,:,:]   )
            
            if (mu_wall.ndim!=3): ## (x,z,t)
                raise ValueError
            if (mu_wall.shape!=(nxr,nz,nt)): ## (x,z,t)
                raise ValueError
            
            if (rho_wall.ndim!=3): ## (x,z,t)
                raise ValueError
            if (rho_wall.shape!=(nxr,nz,nt)): ## (x,z,t)
                raise ValueError
            
            tau_uy = np.copy( mu_wall * ddy_u_wall ) ## INSTANTANEOUS τw
            tau_vy = np.copy( mu_wall * ddy_v_wall )
            tau_wy = np.copy( mu_wall * ddy_w_wall )
            
            u_tau = np.copy( np.sign(tau_uy) * np.sqrt( np.abs(tau_uy) / rho_wall ) ) ## INSTANTANEOUS uτ
            v_tau = np.copy( np.sign(tau_vy) * np.sqrt( np.abs(tau_vy) / rho_wall ) ) ## INSTANTANEOUS vτ
            w_tau = np.copy( np.sign(tau_wy) * np.sqrt( np.abs(tau_wy) / rho_wall ) ) ## INSTANTANEOUS wτ
            
            #grp      = self['data']
            #scalars  = list(grp.keys())
            scalars = [
                    'tau_uy','tau_vy','tau_wy',
                    'u_tau','v_tau','w_tau',
                    'T','mu','nu','rho','p',
                    ]
            
            ## 3D [scalar][x,z,t] numpy structured array
            ## rank-local data buffer for write
            formats  = [ np.dtype(np.float32) for s in scalars ] ## casting to single
            data4spd = np.zeros(shape=(nxr,hf_eas4.nz,hf_eas4.nt), dtype={'names':scalars, 'formats':formats})
            data4spd['tau_uy'][:,:,:] = tau_uy   / ( self.rho_inf * self.U_inf**2 )
            data4spd['tau_vy'][:,:,:] = tau_vy   / ( self.rho_inf * self.U_inf**2 )
            data4spd['tau_wy'][:,:,:] = tau_wy   / ( self.rho_inf * self.U_inf**2 )
            data4spd['T'][:,:,:]      = T_wall   / self.T_inf
            data4spd['rho'][:,:,:]    = rho_wall / self.rho_inf
            data4spd['p'][:,:,:]      = p_wall   / ( self.rho_inf * self.U_inf**2 )
            data4spd['mu'][:,:,:]     = mu_wall  / self.mu_inf
            data4spd['nu'][:,:,:]     = nu_wall  / self.nu_inf
            data4spd['u_tau'][:,:,:]  = u_tau    / self.U_inf
            data4spd['v_tau'][:,:,:]  = v_tau    / self.U_inf
            data4spd['w_tau'][:,:,:]  = w_tau    / self.U_inf
            
            tiA = tii
            tiB = tiA + nt
            tii += nt ## increment tii by this EAS4's nt
            
            ## write
            self.comm.Barrier()
            t_start = timeit.default_timer()
            
            if not threeD:
                
                #for scalar in data4spd.dtype.names:
                for scalar in scalars:
                    dset = self[f'data/{scalar}']
                    with dset.collective:
                        dset[rx1:rx2,:,tiA:tiB] = data4spd[scalar][:,:,:]
            
            else: ## Write as chunkless dset per-file
                
                for scalar in scalars:
                    dsn   = f'data/{i_eas4:d}/{scalar}'
                    shape = ( self.nx, self.nz, nt ) ## 'nt' here is from EAS4
                    dtype = np.dtype(np.float32)
                    
                    dset = self.create_dataset(
                                            dsn,
                                            shape=shape,
                                            dtype=dtype,
                                            chunks=None,
                                            )
                    
                    with dset.collective: ## do actual collective write
                        dset[rx1:rx2,:,:] = data4spd[scalar][:,:,:]
            
            self.comm.Barrier()
            t_delta = timeit.default_timer() - t_start
            
            ## Report write speed
            data_gb = self.n_ranks * data4spd.nbytes / 1024**3
            if verbose:
                msg = even_print('write', f'{data_gb:0.2f} [GB]  {t_delta:0.2f} [s]  {(data_gb/t_delta):0.3f} [GB/s]', s=True)
                tqdm.write(msg)
            
            ## Delete source file (under restrictive conditions)
            if delete_after_import:
                if os.path.isfile('do_delete.txt'):
                    if (self.rank==0):
                        tqdm.write(even_print('deleting', fn_eas4, s=True))
                        os.remove(fn_eas4)
                    self.comm.Barrier()
            
            if verbose: progress_bar.update()
        if verbose: progress_bar.close()
    
    # ==============================================================
    # MOVE from 3D, chunkless HDF5 file to CHUNKED SPD file
    # ==============================================================
    
    if not init_dsets_only and fn_h5_3D is not None:
        
        # ## report contents of 3D file
        # if (self.rank==0):
        #     with h5py.File(fn_h5_3D, 'r') as hfr:
        #         h5_print_contents(hfr)
        # if self.usingmpi:
        #     self.comm.Barrier()
        
        ## Get 'file indices' -- 'indices' are digits in 'data/%d/...' dset names
        if (self.rank==0):
            fi = []
            with h5py.File(fn_h5_3D, 'r') as hfr:
                gpd = hfr['data']
                fi_list = sorted([int(name) for name in gpd.keys() if name.isdigit()])
                nfi_actual = len([ fi for fi in fi_list if fi >= fi_min ]) ## n files to actually process
                fi = np.array(fi_list, dtype=np.int32)
        else:
            fi = np.array([], dtype=np.int32) ## 'fi' must exist on all ranks before bcast
        if self.usingmpi:
            self.comm.Barrier()
            fi = self.comm.bcast(fi, root=0)
        
        if fi.shape[0]==0:
            raise RuntimeError
        
        nx = self.nx
        #ny = self.ny ## doesnt exist
        nz = self.nz
        #nt = self.nt ## no!
        
        scalars = [
                'tau_uy','tau_vy','tau_wy',
                'u_tau','v_tau','w_tau',
                'T','mu','nu','rho','p',
                ]
        
        if verbose:
            progress_bar = tqdm(
                #total=fi.shape[0]*len(scalars),
                total=nfi_actual*len(scalars),
                ncols=100,
                desc='import',
                leave=True,
                file=sys.stdout,
                mininterval=1/24,
                smoothing=0.3,
                #bar_format="\033[B{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\033[A\n\b",
                bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
                ascii="░█",
                colour='#FF6600',
                )
        
        tii = 0 ## timestep index counter full series
        
        with h5py.File(fn_h5_3D, 'r', driver=self.driver, comm=self.comm) as hfr:
            for fii in fi:
                for scalar in scalars:
                    
                    ## dataset name & handle in HDF5 file composed of chunkless dsets
                    dsn  = f'data/{fii:d}/{scalar}'
                    dset = hfr[dsn]
                    nt   = dset.shape[2] ## nt of corresponding chunkless dset (EAS4)
                    
                    ## time range for this EAS4 data
                    tiA = tii
                    tiB = tiA + nt
                    #if verbose:
                    #    tqdm.write(f'tiA={tiA:d},tiB={tiB:d}')
                    
                    if fii >= fi_min: ## do read / write
                        
                        data_gb = nx * nz * nt * dset.dtype.itemsize / 1024**3
                        
                        ## COLLECTIVE read
                        if self.usingmpi: self.comm.Barrier()
                        t_start = timeit.default_timer()
                        if self.usingmpi:
                            with dset.collective:
                                dd = np.copy( dset[rx1:rx2,:,:] )
                        else:
                            dd = np.copy( dset[()] )
                        if self.usingmpi: self.comm.Barrier()
                        t_delta = timeit.default_timer() - t_start
                        
                        if verbose:
                            tqdm.write(even_print(f'read: {fii:d}/{scalar}', '%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
                        
                        ## assert shape
                        if ( dd.shape != (nxr,nz,nt) ):
                            print(f'rank {self.rank:d}: shape violation')
                            if self.usingmpi: self.comm.Abort(1)
                            raise ValueError
                        
                        ## dataset name & handle in SPD file
                        dsn  = f'data/{scalar}'
                        dset = self[dsn]
                        
                        ## COLLECTIVE write
                        if self.usingmpi: self.comm.Barrier()
                        t_start = timeit.default_timer()
                        if self.usingmpi:
                            with dset.collective:
                                dset[rx1:rx2,:,tiA:tiB] = dd[:,:,:]
                        else:
                            dset[:,:,tiA:tiB] = dd[:,:,:]
                        if self.usingmpi: self.comm.Barrier()
                        t_delta = timeit.default_timer() - t_start
                        
                        if verbose:
                            tqdm.write(even_print(f'write: data/{scalar}', '%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
                        if verbose:
                            progress_bar.update()
                    
                    if os.path.isfile('stop.txt'):
                        break
                
                tii += nt ## increment tii by this EAS4's nt
                
                if os.path.isfile('stop.txt'):
                    break
            
            if verbose:
                progress_bar.close()
    
    ## Report file
    if self.usingmpi:
        self.comm.Barrier()
    if verbose:
        print(72*'-')
        even_print( os.path.basename(self.fname), f'{(os.path.getsize(self.fname)/1024**3):0.1f} [GB]')
    if verbose: print(72*'-')
    if verbose: print('total time : spd.import_eas4_wall() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return



import numpy as np
from scipy.integrate import cumulative_trapezoid
from scipy.optimize import root_scalar
from scipy.special import exp1

# ======================================================================

def asymptotic_Wplus_eta(eta, y_plus, Re_theta):
    '''
    Monkewitz et al. (2007)
    Full-boundary-layer wake function
    -----
    This function generates a Re-dependent full-layer W+
      (not just W+_outer) using the U+ composite profile
    '''
    kappa = 0.384
    Cstar = 3.354
    
    ## Shape factor
    H12 = asymptotic_H12_Retheta(Re_theta)
    
    ## Coles-Fernholz
    ue_plus = (1./kappa)*np.log(H12*Re_theta) + Cstar
    
    ## Composite velocity
    uplus_composite = asymptotic_uplus_composite(y_plus, eta, Re_theta)
    
    ## Wake (velocity defect)
    return ue_plus - uplus_composite

def asymptotic_uplus_inner(y_plus):
    '''
    Monkewitz et al. (2007)
    Inner mean velocity U^+_{inner}(y^+)
    Obtained by integrating the Padé indicator function
    Eqs. (5)-(7)
    '''
    y_plus = np.asarray(y_plus)
    
    k = 0.384
    
    ## Padé P23 coefficients (Eq. 5)
    b0 = 1.000e-2 / k
    b1 = 1.100e-2
    b2 = 1.100e-4
    
    P23 = b0 * (1. + b1*y_plus + b2*y_plus**2) / (
          1. + b1*y_plus + b2*y_plus**2 + k*b0*b2*y_plus**3
          )
    
    ## Padé P25 coefficients (Eq. 6)
    h1 = -1.000e-2
    h2 =  6.000e-3
    h3 =  9.977e-4
    h4 =  2.200e-5
    h5 =  1.000e-6
    
    P25 = (1. - b0) * (1. + h1*y_plus + h2*y_plus**2) / (
          1. + h1*y_plus + h2*y_plus**2 + h3*y_plus**3 + h4*y_plus**4 + h5*y_plus**5
          )
    
    d_uplus_d_yplus = P23 + P25 ## d[u+]/dy+
    
    ## scipy.integrate.quad() could be used for continuous integration
    
    ## discrete trapezoidal integration
    uplus_inner = cumulative_trapezoid(y=d_uplus_d_yplus, x=y_plus, initial=0.)
    return uplus_inner

def asymptotic_uplus_composite(y_plus, eta, Re_theta):
    '''
    Monkewitz et al. (2007)
    Composite mean velocity U^+(y^+,η)
    Eq. (15), with U_e^+ from H12(Re_θ)
    '''
    y_plus = np.asarray(y_plus)
    eta    = np.asarray(eta)
    n      = y_plus.shape[0]
    
    kappa  = 0.384
    B      = 4.17   ## log-law intercept
    Cstar  = 3.354  ## C* (Monkewitz / Nagib)
    
    ## Inner velocity (Padé-based)
    uplus_inner = asymptotic_uplus_inner(y_plus)
    
    ## Shape factor and edge velocity
    H12       = asymptotic_H12_Retheta(Re_theta)
    ue_plus = (1.0/kappa)*np.log(H12*Re_theta) + Cstar
    
    ## Outer velocity
    Wplus_outer = asymptotic_Wplus_eta_outer(eta)
    uplus_outer = ue_plus - Wplus_outer
    
    ## Overlap (log law)
    uplus_log     = np.full((n,),0.,dtype=np.float64)
    uplus_log[1:] = (1.0/kappa)*np.log(y_plus[1:]) + B
    uplus_log[0]  = B
    
    ## Additive composite
    uplus_composite = uplus_inner + uplus_outer - uplus_log
    uplus_composite[0] = 0.
    
    return uplus_composite

def asymptotic_Wplus_eta_outer(eta):
    '''
    Monkewitz et al. (2007)
    https://doi.org/10.1063/1.2780196
    Eq.9
    '''
    ny  = eta.shape[0]
    k   = 0.384 ## κ
    w0  = 0.6332
    wn1 = -0.096 ## w_{-1}
    w2  = 28.5
    w8  = 33000.
    
    Wplus_outer = np.full((ny,),np.nan,dtype=np.float64)
    
    ii = np.where(eta>0)[0]
    η = eta[ii]
    
    ## W_+ = ((1/κ) E1(η) + w0) * 0.5 * [1 - tanh(w_{-1}/η + w2 η^2 + w8 η^8)]
    Wplus_outer[ii] = ( (1./k)*exp1(eta[ii]) + w0 ) \
                      * 0.5 \
                      * ( 1. - np.tanh( (wn1/η) + w2*η**2 + w8*η**8 ) )
    
    return Wplus_outer

def asymptotic_H12_Retheta(Re_theta):
    '''
    H12 = δ1/δ2 = δ*/θ
    Reθ = θ·ue·ρe/μe
    -----
    Monkewitz et al. (2007)
    https://doi.org/10.1063/1.2780196
    Nagib et al. (2007)
    https://doi.org/10.1098/rsta.2006.1948
    '''
    
    ## Asymptotic constants
    Cs = 3.354 ## C*
    k  = 0.384 ## κ
    I2 = 7.135 ## Iww (Monkewitz) | C′ (Nagib)
    
    Re_theta = np.atleast_1d(np.asarray(Re_theta, dtype=np.float64))
    n = Re_theta.shape[0]
    H12 = np.full((n,),np.nan,dtype=np.float64)
    
    def __f_root_nonlin_H12(H12,Re_theta):
        ue_plus = (1/k)*np.log(H12*Re_theta) + Cs ## Coles-Fernholz
        H12_test = 1. / ( 1. - (I2/ue_plus) )
        root = H12_test - H12
        return root
    
    def __f_H12_Retheta(Re_theta):
        n = Re_theta.shape[0]
        H12_out = np.zeros((n,),dtype=np.float64)
        for i in range(n):
            sol = root_scalar(
                    f=__f_root_nonlin_H12,
                    args=(Re_theta[i],),
                    x0=1.35,
                    method='brentq',
                    bracket=[0.99,3.01],
                    )
            if not sol.converged:
                raise ValueError
            H12_out[i] = float(sol.root)
        
        return H12_out
    
    ## Main solution
    H12 = __f_H12_Retheta(Re_theta)
    
    ## Final check
    ue_plus = (1/k)*np.log(H12*Re_theta) + Cs ## Coles-Fernholz
    H12_test = 1. / ( 1. - (I2/ue_plus) )
    np.testing.assert_allclose(H12_test, H12, rtol=1e-6, atol=1e-8)
    
    if n==1:
        return float(H12[0])
    
    return H12

def asymptotic_H32_Retheta(Re_theta,I3):
    '''
    H32 = δ3/δ2
    Reθ = θ·ue·ρe/μe
    '''
    
    ## Asymptotic constants (same as H12 function)
    Cs  = 3.354 ## C*
    k   = 0.384 ## κ
    I2  = 7.135 ## Iww (Monkewitz) | C′ (Nagib)
    
    if isinstance(Re_theta,(int,np.int32,np.int64,float,np.float32,np.float64)):
        Re_theta = np.array([Re_theta,],np.float64)
    elif isinstance(Re_theta, np.ndarray):
        pass
    elif isinstance(Re_theta, list):
        Re_theta = np.array(Re_theta,np.float64)
    else:
        raise ValueError
    
    ## H12(Reθ)
    H12 = asymptotic_H12_Retheta(Re_theta)
    
    if isinstance(H12,float):
        H12 = np.array([H12,],np.float64)
    
    n = Re_theta.shape[0]
    H32 = np.zeros((n,),dtype=np.float64)
    
    for i in range(n):
        ue_plus = (1/k)*np.log(H12[i]*Re_theta[i]) + Cs ## Coles–Fernholz
        H32[i] = (2. - 3.*(I2/ue_plus) + I3/(ue_plus**2)) / (1. - (I2/ue_plus))
    
    if H32.shape[0]==1:
        H32 = float(H32)
    
    return H32


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
========================================================================
Command line interface
========================================================================
'''

import argparse


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input", help="Help message for input")
    args = parser.parse_args()
    print(args.input)
    ## ... this is a dummy CLI for now
    ## call from cmd line with e.g. >$ turbx 'hello'



import numpy as np
from scipy.integrate import cumulative_trapezoid

from .gradient import gradient

# ======================================================================

def comp_transform_TL(y,u,utau,rho,mu,**kwargs):
    '''
    Trettel & Larsson (2016)
    https://doi.org/10.1063/1.4942022
    '''
    acc          = kwargs.get('acc',6)
    edge_stencil = kwargs.get('edge_stencil','full')
    ddy_rho      = kwargs.get('ddy_rho',None)
    ddy_mu       = kwargs.get('ddy_mu',None)
    
    ## Calculate if not passed
    if ddy_rho is None:
        ddy_rho = gradient(rho, y, acc=acc, edge_stencil=edge_stencil)
    if ddy_mu is None:
        ddy_mu = gradient(mu, y, acc=acc, edge_stencil=edge_stencil)
    
    rho_w     = rho[0]
    mu_w      = mu[0]
    u_plus    = u / utau
    tau_wall  = rho_w * utau**2
    sc_l_in   = mu_w / ( utau * rho_w ) ## δν
    #y_plus    = y * utau * rho_w / mu_w
    
    ## Eq.28
    bracket   = 1. + ((1./(2.*rho))*ddy_rho*y) - ((1/mu)*ddy_mu*y)
    integrand = np.sqrt(rho/rho_w) * bracket
    uTL_plus  = cumulative_trapezoid(y=integrand, x=u_plus, initial=0.) ## U+
    
    u_sl = np.sqrt( tau_wall / rho ) ## semilocal friction velocity (NOT transformed U+)
    y_sl = y * u_sl * rho / mu       ## semilocal coordinate
    
    ## Re-dimensionalize first two returned profiles by uτ,δν
    return uTL_plus*utau, y_sl*sc_l_in, uTL_plus, y_sl

def comp_transform_VD(u,rho):
    '''
    van Driest velocity transform
    uVD(y) = ∫ √(ρ/ρ_w) du
    van Driest (1951)
    https://doi.org/10.2514/8.1895
    '''
    rho_w     = rho[0]
    integrand = np.sqrt(rho/rho_w)
    uVD       = cumulative_trapezoid(y=integrand, x=u, initial=0.)
    return uVD

def comp_transform_VIPL(y,u,rho,mu,a=1.5,b=0.5):
    '''
    Volpiani et al. (2020)
    https://doi.org/10.1103/PhysRevFluids.5.052602
    '''
    rho_w = rho[0]
    mu_w  = mu[0]
    M     = mu/mu_w
    R     = rho/rho_w
    F     = R**b * M**(-a)
    yVP   = cumulative_trapezoid(y=F, x=y, initial=0.)
    G     = R**b * M**(1-a)
    uVP   = cumulative_trapezoid(y=G, x=u, initial=0.)
    return uVP,yVP

def comp_transform_GFM(y,uFv,rho,mu,r_uII_vII,utau,ddy_u_Fv=None,acc=6,edge_stencil='full'):
    '''
    Griffin, Fu & Moin (2021)
    https://doi.org/10.1073/pnas.2111144118
    '''
    if ddy_u_Fv is None:
        ddy_u_Fv = gradient(uFv, y, acc=acc, edge_stencil=edge_stencil)
    
    ny       = y.shape[0]
    rho_w    = rho[0]
    mu_w     = mu[0]
    tau_wall = utau**2 * rho_w
    sc_l_in  = mu_w / ( utau * rho_w ) ## δν
    y_plus   = y * utau * rho_w / mu_w
    uFv_plus = uFv / utau
    mu_plus  = mu  / mu_w
    
    tau_visc_plus = mu * ddy_u_Fv / tau_wall    ## τv+
    tau_R_plus    = -1. * r_uII_vII / tau_wall ## τR+
    tau_plus      = tau_visc_plus + tau_R_plus ## τ+
    
    u_sl = np.sqrt( tau_wall / rho ) ## u_{sl} : semilocal friction velocity
    #sc_l_sl = mu / ( u_sl * rho )    ## ℓ_{sl} : semilocal length scale
    y_star  = y * u_sl * rho / mu    ## y* = y_{sl} = y/ℓ_{sl} : semilocal coordinate
    
    duFvplus_dyplus = gradient(uFv_plus, y_plus, acc=acc, edge_stencil=edge_stencil)
    duFvplus_dystar = gradient(uFv_plus, y_star, acc=acc, edge_stencil=edge_stencil)
    
    #duFvplus_dyplus = ddy_u_Fv / (utau / sc_l_in) ## ∂[uFv+]/∂y+ -- OK, sc_l_in is a constant over [y]
    #duFvplus_dystar = ddy_u_Fv / (utau / sc_l_sl) ## ∂[uFv+]/∂y* -- WRONG, sc_l_sl is NOT constant over [y]
    
    # np.testing.assert_allclose(
    #     duFvplus_dyplus,
    #     gradient(uFv_plus, y_plus, acc=acc, edge_stencil=edge_stencil),
    #     rtol=1e-6,
    #     atol=1e-12,
    #     )
    
    # np.testing.assert_allclose(
    #     duFvplus_dystar,
    #     gradient(uFv_plus, y_star, acc=acc, edge_stencil=edge_stencil),
    #     rtol=1e-6,
    #     atol=1e-12,
    #     )
    
    S_eq_plus = (1./mu_plus) * duFvplus_dystar ## Eq.1
    S_TL_plus = mu_plus * duFvplus_dyplus      ## Eq.2
    
    ## Eq.4 St+ = [τ+·Seq+] / [ τ+ + Seq+ - STL+ ]
    A = tau_plus * S_eq_plus
    B = tau_plus + S_eq_plus - S_TL_plus
    B = np.where(np.abs(B) < 1e-12, 1e-12, B)
    S_t_plus = A / B
    
    ## Eq.3 Assertion
    if False:
        ratio_v = np.full((ny,),0.,dtype=np.float64)
        ratio_R = np.full((ny,),0.,dtype=np.float64)
        np.divide(tau_visc_plus, S_TL_plus, out=ratio_v, where=np.abs(S_TL_plus) > 1e-12)
        np.divide(tau_R_plus,    S_eq_plus, out=ratio_R, where=np.abs(S_eq_plus) > 1e-12)
        tau_plus_chk = S_t_plus * (ratio_v + ratio_R)
        mask = ( (np.abs(S_TL_plus) > 1e-12) & (np.abs(S_eq_plus) > 1e-12) )
        np.testing.assert_allclose(
            tau_plus[mask],
            tau_plus_chk[mask],
            rtol=1e-6,
            )
    
    ## Ut+[y*] = ∫St+ dy*
    uGFM_plus = cumulative_trapezoid(y=S_t_plus, x=y_star, initial=0.)
    
    ## Re-dimensionalize first two returned profiles by uτ,δν
    return uGFM_plus*utau, y_star*sc_l_in, uGFM_plus, y_star
